#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level logging handler subclasses.
'''

#FIXME: Transparently compress rotated logfiles with the standard ".gz"-style
#compression format. (Surely, StackOverflow has already solved this.)

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid circular import dependencies, avoid importing from *ANY*
# application-specific modules at the top-level -- excluding those explicitly
# known *NOT* to import from this module. Since all application-specific
# modules must *ALWAYS* be able to safely import from this module at any level,
# these circularities are best avoided here rather than elsewhere.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import errno, time
from betse.exceptions import BetseLogRaceException
from betse.util.io import stderrs
from logging import LogRecord
from logging.handlers import RotatingFileHandler

# ....................{ SUBCLASSES                        }....................
class LogHandlerFileRotateSafe(RotatingFileHandler):
    '''
    Process-safe rotating file handler.

    The standard :class:`RotatingFileHandler` class is thread- but *not*
    process-safe. Concurrent attempts to log to the same physical file from
    multiple processes can and typically will produce fatal race conditions
    producing raised exceptions from one or more of these processes. On
    logfile rotation, each process will aggressively contend with each other
    process for write access to the same physical file to be rotated.

    This :class:`RotatingFileHandler` subclass is both thread- *and*
    process-safe, obviating these concerns.

    Caveats
    ----------
    This handler does *not* implement platform-specific file locking to
    constrain logfile access during rotation. While feasible (e.g., via the
    third-party ConcurrentLogHandler_ package), doing so is complicated by
    numerous non-trivial tradeoffs -- including:

    * **Efficiency.** File locking is notoriously inefficient. File locking
      logfile access, which necessitates acquiring and release a file lock for
      each log record written to disk by this handler, is doubly so. Ideally,
      logging should have negligible to no performance implications. When this
      ceases to be the case, logging ceases to be universally viable.
    * **Security.** File locking is notoriously insecure. Both BSD-style
      ``flock()`` locks *and* POSIX-style ``fcntl()`` locks on files having
      permissions more permissive than `0600` (e.g., group- or world-readable)
      expose applications to permanent deadlocks. Malicious users with group-
      or world-readable access to files to be locked can permanently halt the
      execution of applications locking those files by:

      #. Preemptively acquiring those locks first.
      #. Permanently preserving those locks (i.e., never releasing those
         locks).

    * **Thread safety.** Specifically:

      * POSIX-style ``fcntl()`` locks lock on process IDs (PIDs) and hence are
        implicitly non-thread-safe.
      * BSD-style ``flock()`` locks lock on file descriptors (FDs) and hence
        are implicitly thread-safe.

    * **Portability.** While the prior point implies BSD-style ``flock()``
      locks to be preferable to POSIX-style ``fcntl()`` locks for purposes of
      thread-safety, the latter remain mildly more portable than the former.
      Naturally, neither are supported under Microsoft platforms, which
      provides entirely different synchronization platforms suffering
      completely different tradeoffs (e.g., inability to rename files
      concurrently opened by multiple processes). Python's stdlib implements
      low-level wrappers encapsulating all three platform-specific APIs but
      *no* high-level wrapper unifying these fundamentally dissimilar
      approaches under a common API. Optional third-party packages implementing
      these high-level wrappers exist but introduce additional tradeoffs (in
      addition to those documented above), including:

      * **Maintenance.** Most packages are poorly maintained at best, with
        stable releases few and far between.
      * **Documentation.** Most packages are poorly documented at best, with
        little to no human-readable documentation published online.
      * **Uncertainty.** While multiple packages exist, no package appears to
        predominate the others with respect to popularity or usage.

    These deficiencies and more are well-documented, long-standing, and
    unlikely to be resolved on any platform at any point in the near future.
    Until universally resolved on *all* supported platforms, the existence of
    these deficiencies implies file locking to be broken by design and unusable
    for applications in the "real world."

    Non-filesystem locking (e.g., :mod:`multiprocessing` module synchronization
    primitives) is strongly preferable. Unfortunately, such locking requires
    centralized coordination managed at the high-level Python level. No such
    coordination exists between independent BETSE processes run by external
    users at the low-level operating system level.

    Both file- and non-file locking are inapplicable within this context. Ergo,
    this handler cannot reasonably constrain logfile access during rotation.
    Instead, on detecting exceptions produced by race conditions between
    multiple processes competing for access when attempting to emit log
    records, this handler temporarily halts the current process for a
    negligible amount of the timeslice (e.g., 100ms) and repeats this attempt a
    negligible number of times (e.g., 8) *before* giving up and raising a fatal
    :exc:`BetseLogRaceException`.

    External callers may manually circumvent these race conditions by passing
    the `--log-file LOG_FILENAME` option at the command line, where
    `LOG_FILENAME` is the absolute filename of a (possibly non-existent) log
    file unique to the current process.

    While non-ideal, no sane solutions exist. File locking is insane.

    .. _ConcurrentLogHandler:
       https://pypi.python.org/pypi/ConcurrentLogHandler

    See Also
    ----------
    http://0pointer.de/blog/projects/locking.html
        *On the Brokenness of File Locking.* Classic treatise on the
        long-standing deficiencies of both BSD- and POSIX-style file locking.
    https://loonytek.com/2015/01/15/advisory-file-locking-differences-between-posix-and-bsd-locks
        *Advisory File Locking – My take on POSIX and BSD locks.* Detailed
        analysis concluding with similar deficiencies and lack of solutions.
    '''

    # ..................{ EMITTERS                          }..................
    def emit(self, record: LogRecord) -> None:
        '''
        Log the passed logging record in a thread- *and* process-safe manner.

        Parameters
        ----------
        record : LogRecord
            Logging record to be logged.

        Raises
        ----------
        BetseLogRaceException
            If this method detects but fails to automatically resolve a logging
            race condition between multiple processes concurrently contending
            for write access to the same logfile.
        '''

        # Attempt to emit this record to this logfile and conditionally rotate
        # this logfile in the default non-process-safe manner.
        try:
            return super().emit(record)
        # If an exception indicative of a race condition between multiple
        # processes competing for logfile rotation is raised, attempt to
        # re-emit this record to this logfile in a process-safe manner. Since
        # this necessarily entails inefficiency, we do so *ONLY* as needed.
        #
        # These exceptions are indicated by the following real-world
        # tracebacks:
        #
        #     # Traceback implicating the "FileNotFoundError" exception.
        #     Traceback (most recent call last):
        #       File "/opt/shared/python/3.6.0/lib/python3.6/logging/handlers.py", line 72, in emit
        #         self.doRollover()
        #       File "/opt/shared/python/3.6.0/lib/python3.6/logging/handlers.py", line 169, in doRollover
        #         os.rename(sfn, dfn)
        #     FileNotFoundError: [Errno 2] No such file or directory: '/cluster/home/sburck01/.betse/betse.log.4' -> '/cluster/home/sburck01/.betse/betse.log.5'
        #     Call stack:
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/bin/betse", line 63, in <module>
        #         sys.exit(entry_module.main())
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/lib/python3.6/site-packages/betse-0.4.2-py3.6.egg/betse/cli/__main__.py", line 45, in main
        #         return CLIMain().run(arg_list)
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/lib/python3.6/site-packages/betse-0.4.2-py3.6.egg/betse/cli/cliabc.py", line 156, in run
        #         matplotlib_backend_name=self._args.matplotlib_backend_name)
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/lib/python3.6/site-packages/betse-0.4.2-py3.6.egg/betse/lib/libs.py", line 258, in reinit
        #         init(*args, **kwargs)
        #       File "<string>", line 15, in func_type_checked
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/lib/python3.6/site-packages/betse-0.4.2-py3.6.egg/betse/lib/libs.py", line 291, in init
        #         mpl_config.init(backend_name=matplotlib_backend_name)
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/lib/python3.6/site-packages/betse-0.4.2-py3.6.egg/betse/lib/matplotlib/matplotlibs.py", line 226, in init
        #         self._init_backend(backend_name=backend_name)
        #
        #     # Traceback implicating the "OSError" exception.
        #     Traceback (most recent call last):
        #       File "/opt/shared/python/3.6.0/lib/python3.6/logging/handlers.py", line 71, in emit
        #         if self.shouldRollover(record):
        #       File "/opt/shared/python/3.6.0/lib/python3.6/logging/handlers.py", line 188, in shouldRollover
        #         self.stream.seek(0, 2)  #due to non-posix-compliant Windows feature
        #     OSError: [Errno 116] Stale file handle
        #     Call stack:
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/bin/betse", line 63, in <module>
        #         sys.exit(entry_module.main())
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/lib/python3.6/site-packages/betse-0.4.2-py3.6.egg/betse/cli/__main__.py", line 45, in main
        #         return CLIMain().run(arg_list)
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/lib/python3.6/site-packages/betse-0.4.2-py3.6.egg/betse/cli/cliabc.py", line 156, in run
        #         matplotlib_backend_name=self._args.matplotlib_backend_name)
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/lib/python3.6/site-packages/betse-0.4.2-py3.6.egg/betse/lib/libs.py", line 258, in reinit
        #         init(*args, **kwargs)
        #       File "<string>", line 15, in func_type_checked
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/lib/python3.6/site-packages/betse-0.4.2-py3.6.egg/betse/lib/libs.py", line 291, in init
        #         mpl_config.init(backend_name=matplotlib_backend_name)
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/lib/python3.6/site-packages/betse-0.4.2-py3.6.egg/betse/lib/matplotlib/matplotlibs.py", line 226, in init
        #         self._init_backend(backend_name=backend_name)
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/lib/python3.6/site-packages/betse-0.4.2-py3.6.egg/betse/lib/matplotlib/matplotlibs.py", line 434, in _init_backend
        #         if self.is_backend_usable(_BACKEND_NAME_HEADLESS):
        #       File "<string>", line 15, in func_type_checked
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/lib/python3.6/site-packages/betse-0.4.2-py3.6.egg/betse/lib/matplotlib/matplotlibs.py", line 760, in is_backend_usable
        #         logs.log_debug('Testing matplotlib backend "%s"...', backend_name)
        #       File "<string>", line 15, in func_type_checked
        #       File "/cluster/kappa/90-days-archive/levin/levinlab/sburck01/venv-betse/lib/python3.6/site-packages/betse-0.4.2-py3.6.egg/betse/util/io/log/logs.py", line 100, in log_debug
        #         logging.debug(message, *args, **kwargs)
        #
        # If this is a "FileNotFoundError", unconditionally retry.
        except FileNotFoundError:
            return self._emit_safely(record)
        # Else if this is an "OSError"...
        except OSError as exception:
            # If this is the same error as in the traceback above, retry.
            if exception.errno == errno.ESTALE:
                return self._emit_safely(record)
            # Else, re-raise this exception.
            else:
                raise
        # Else, permit this exception to continue unwinding the call stack.

    # ..................{ PRIVATE                           }..................
    # Note that, while the emit() method defined above *COULD* be reimplemented
    # to perform the iteration performed by this method and this method then
    # removed, doing so would inefficiently incur the cost of such iteration on
    # every logging call -- which is clearly unacceptable.
    #
    # Instead, we bite the DRY bullet and simply repeat ourselves below.
    def _emit_safely(self, record: LogRecord) -> None:
        '''
        Attempt to repeatedly emit the passed logging record in a process-safe
        manner *after* the parent :meth:`emit` call fails to do so.

        Parameters
        ----------
        record : LogRecord
            Logging record to be logged.

        Raises
        ----------
        BetseLogRaceException
            If this method detects but fails to automatically resolve a logging
            race condition between multiple processes concurrently contending
            for write access to the same logfile.
        '''

        # Arbitrary maximum number of times to attempt to re-emit this record.
        ATTEMPTS_MAX = 8

        # Number of seconds to temporarily halt this process, equal to 100ms.
        SLEEP_INTERVAL = 0.1

        # For each such attempt...
        for attempt_index in range(ATTEMPTS_MAX):
            # Notify the user of this race condition. Due to the circumstances,
            # logging this message is right out.
            stderrs.output(
                'Detected logging race condition between '
                'multiple processes...')
            stderrs.output(
                'Sleeping for {}s before retrying {}/{} times...'.format(
                    SLEEP_INTERVAL, attempt_index, ATTEMPTS_MAX))

            # Temporarily halt the current process in the hope that this race
            # condition will be resolved before this process is awakened.
            time.sleep(SLEEP_INTERVAL)

            # Attempt to re-emit this record to this logfile and conditionally
            # rotate this logfile in the default non-process-safe manner,
            # immediately returning on success.
            try:
                return super().emit(record)
            # Catch the same exceptions caught by the emit() method above.
            #
            # If this is a "FileNotFoundError", unconditionally retry.
            except FileNotFoundError:
                continue
            # Else if this is an "OSError"...
            except OSError as exception:
                # If this is the same error as in the traceback above, retry.
                if exception.errno == errno.ESTALE:
                    continue
                # Else, re-raise this exception.
                else:
                    raise
            # Else, permit this exception to continue unwinding the call stack.

        # Attempt to re-emit this record to this logfile one last time...
        try:
            return super().emit(record)
        # If this attempt fails, raise a human-readable application-specific
        # exception chained onto this presumably non-human-readable exception.
        except Exception as exception:
            raise BetseLogRaceException(
                'Fatal logging race condition between multiple processes '
                'detected after {} failed attempts at automatic resolution. '
                'External callers may manually circumvent this issue by '
                'passing the "--log-file LOG_FILENAME" option at the '
                'command line, where "LOG_FILENAME" is the absolute filename '
                'of a (possibly non-existent) log file unique to the current '
                'process.'.format(ATTEMPTS_MAX)
            ) from exception



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests exercising the
:class:`betse.util.type.iterable.mapping.mapcls.DynamicValueType` type.
'''

# ....................{ IMPORTS                           }....................
import pytest
import sys
from pytest import fixture

# ....................{ GLOBALS                           }....................
# Global variables whose values are referenced by the dictionary returned by
# the "betse_dynamicvaluedict" fixture.

module = sys.modules[__name__]
'''
Current module object, for use in :func:`setattr` calls below.
'''


varda = (
    'O stars that in the Sunless Year',
    'With shining hand by her were sown,',
    'In Windy fields now bright and clear',
    'We see your silver blossom blown!'
)

yavanna = (
    "'All have their worth,' said Yavanna,",
    "'and each contributes to the worth of the others.",
    "But the kelvar can flee or defend themselves,"
    "whereas the olvar that grow cannot.",
    "And among these I hold trees dear.",
    "Long in the growing, swift shall they be in the felling,",
    "and unless they pay toll with fruit upon bough",
    "little mourned in their passing.",
    "So I see in my thought.",
    "Would that the trees might speak on behalf of all things that have roots,",
    "and punish those that wrong them!'",
)

# ....................{ FIXTURES                          }....................
@fixture(scope='session')
def betse_dynamicvaluedict() -> 'DynamicValueDict':
    '''
    Fixture creating and returning a mock :class:`DynamicValueDict` to be
    tested.
    '''

    # Defer heavyweight imports.
    from betse.util.type.iterable.mapping.mapcls import (
        DynamicValue, DynamicValueDict)

    # Create and return an instance of this type.
    return DynamicValueDict({
        'Varda Elentári': DynamicValue(
            lambda: varda,

            # Sadly, lambda expressions prohibit assignment in Python. *URGH!*
            lambda value: setattr(module, 'varda', value),
            # lambda value: sim.cc_cells.__setindex__(sim.iNa, value)
        ),
        'Yavanna': DynamicValue(
            lambda: yavanna,
            lambda value: setattr(module, 'yavanna', value),
        ),
    })

# ....................{ TESTS                             }....................
def test_dynamicvaluedict_pass(betse_dynamicvaluedict) -> None:
    '''
    Test all aspects of the :class:`DynamicValueDict` type intended to succeed.

    Parameters
    ----------
    betse_dynamicvaluedict : DynamicValueDict
        Instance of this type to be tested.
    '''

    # Test this type's __len__() implementation.
    assert len(betse_dynamicvaluedict) == 2

    # Test this type's __iter__() implementation.
    for dynamicvalue in betse_dynamicvaluedict:
        assert isinstance(dynamicvalue, str)

    # Test this type's __getindex__() implementation.
    assert len(betse_dynamicvaluedict['Varda Elentári']) == 4

    # Test this type's __setindex__() implementation. Specifically, validate
    # that setting a dictionary value sets the value of the global variable
    # underlying this value.
    assert len(yavanna) == 10
    betse_dynamicvaluedict['Yavanna'] = (
        "'Yet it was in the Song,' said Yavanna.",
        "'For while thou wert in the heavens",
        "and with Ulmo built the clouds and poured out the rains,",
        "I lifted up the branches of great trees to receive them,",
        "and some sang to Ilúvatar amid the wind and the rain.'",
    )
    assert len(yavanna) == 5


def test_dynamicvaluedict_fail(betse_dynamicvaluedict) -> None:
    '''
    Test all aspects of the :class:`DynamicValueDict` type intended to fail.

    Parameters
    ----------
    betse_dynamicvaluedict : DynamicValueDict
        Instance of this type to be tested.
    '''

    # Imports deferred for safety.
    from betse.exceptions import BetseMethodUnimplementedException

    # Test this type's __setindex__() implementation. Specifically, validate
    # that setting a dictionary value *NOT* predefined at dictionary
    # initialization time fails.
    with pytest.raises(KeyError):
        betse_dynamicvaluedict['Estë'] = (
            'Grey is her raiment,',
            'and rest her gift.',
        )

    # Test this type's __delindex__() implementation.
    with pytest.raises(BetseMethodUnimplementedException):
        del betse_dynamicvaluedict['Varda Elentári']



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **character** (i.e., string of length 1) facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseCharException
from betse.util.type.types import type_check

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_unless_char(text: str) -> None:
    '''
    Raise an exception unless the passed string is a **character** (i.e., of
    length 1).

    Parameters
    ----------
    text : str
        String to be validated.

    Raises
    ----------
    BetseCharException
        If this string is either:
        * The empty string.
        * Of length greater than 1.
    '''

    # If this string is *NOT* a character, raise an exception.
    if not is_char(text):
        raise BetseCharException(
            'String "{}" not a character '
            '(i.e., either empty or of length >= 2).'.format(text))

# ....................{ TESTERS ~ case                    }....................
@type_check
def is_char(text: str) -> bool:
    '''
    ``True`` only if the passed string is a **character** (i.e., of length 1).

    Parameters
    ----------
    text : str
        String to be tested.

    Returns
    ----------
    bool
        ``True`` only if this string is a character.
    '''

    # Don't judge us.
    return len(text) == 1



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

"""
The toolbox module contains a number of functions that are used throughout the
BETSE project.
"""

# ....................{ IMPORTS                            }....................
import math, copy
import numpy as np
import scipy.spatial as sps
from betse.util.type.types import type_check, SequenceTypes
from scipy import interpolate as interp

# ....................{ UTILITIES                          }....................
#FIXME: Consider shifting this general-purpose sequence method to
#betse.util.type.iterable.sequences.flatten().
@type_check
def flatten(ls_of_ls: SequenceTypes) -> tuple:
    '''
    Flattens a doubly-nested sequences of sequences (e.g., list of lists,
    two-dimensional Numpy array).

    For convenience, this function silently flattens two-dimensional Numpy
    arrays to lists of lists. Notheless, for both sanity and clarity, callers
    are recommended to instead call the standard :meth:`numpy.ndarray.flatten`
    method when this sequence is known to be a Numpy array.

    Parameters
    ----------
    ls_of_ls : SequenceTypes
        Nested sequence of sequences (e.g., ``[[a,b,c],[d,e,f],[g,h,i]]``).

    Returns
    -------
    (list, list, list)
        3-tuple ``(ls_flat, ind_map, rind_map)``, where:
        * ``ls_flat`` is a flattened version of the input list (e.g.,
          ``[a,b,c,d,e,f,g,h,i]``).
        * ``ind_map`` is a list of forward indices mapping from all indices of
          the output ``ls_flat`` list to the corresponding indices of the input
          ``ls_of_ls`` list (e.g., ``ind_map[5] = [0,5]``, which would yield the
          same value for ``ls_flat[5]`` and ``ls_of_ls[0][5]``).
        * ``rind_map`` is a list of lists of reverse indices mapping from all
          indices of the input ``ls_of_ls`` list to the corresponding indices of
          the output ``ls_flat`` list.
    '''

    ls_flat = []
    ind_map = []
    rind_map = copy.deepcopy(ls_of_ls)   # make a deepcopy of the nested list to get the right shape

    for i, sublist in enumerate(ls_of_ls):    # flatten the array and make a mapping: inds nest --> inds flat
        for j, val in enumerate(sublist):
            ls_flat.append(val)
            ind_map.append([i,j])

    for j, vals in enumerate(ind_map):   # go through the mapping and create a reverse mapping: ind flat --> inds nest
        rind_map[vals[0]][vals[1]] = j

    return ls_flat, ind_map, rind_map   # return the flattened list and the map and reverse map

def area(p):
    """
    Calculates the unsigned area of an arbitrarily shaped polygon defined by a set of
    counter-clockwise oriented points in 2D.

    Parameters
    ----------
    p               xy list of polygon points

    Returns
    -------
    area            area of a polygon in square meters

    Notes
    -------
    The algorithm is an application of Green's theorem for the functions -y and
    x, exactly in the way a planimeter works.
    """

    foo = np.asarray(p)

    # move points along by one:
    foo_p = np.roll(foo, -1, axis=0)

    ai = foo[:,0]*foo_p[:,1] - foo_p[:,0]*foo[:,1]

    aa = np.abs((1/2)*np.sum(ai))


    return aa


def poly_centroid(p):
    """
    Calculates the centroid (geometric centre of mass) of a polygon.

    Parameters
    ----------
    p       array of [x,y] points defining polygon vertices

    Returns
    --------
    cx, cy  polygon centroid coordinates

    reference: https://en.wikipedia.org/wiki/Centroid#Of_a_polygon
    """

    # cx, cy = np.mean(p, axis = 0)


    foo = np.asarray(p)

    # move points along by one:
    foo_p = np.roll(foo, -1, axis=0)

    ai = foo[:,0]*foo_p[:,1] - foo_p[:,0]*foo[:,1]

    aa = (1/2)*np.sum(ai)  # signed area

    cx = (1/(6*aa))*np.sum((foo[:,0] + foo_p[:,0])*(foo[:,0] * foo_p[:,1] - foo_p[:,0] * foo[:,1]))
    cy = (1/(6*aa))*np.sum((foo[:,1] + foo_p[:,1])*(foo[:,0] * foo_p[:,1] - foo_p[:,0] * foo[:,1]))


    return cx, cy


def circumc(A, B, C):
    """
    Calculates the circumcenter and circumradius of a triangle with
    vertices A = [Ax, Ay], B = [Bx, By], and C = [Cx, Cy]

    returns ox, oy, rc, the x and y coordinates of the circumcentre and
    the circumradius, respectively.

    """

    # Point coords:
    Ax = A[0]
    Ay = A[1]
    Bx = B[0]
    By = B[1]
    Cx = C[0]
    Cy = C[1]

    # Calculate circumcentre:
    # (from https://en.wikipedia.org/wiki/Circumscribed_circle#Cartesian_coordinates_2)

    A2 = Ax**2 + Ay**2
    B2 = Bx**2 + By**2
    C2 = Cx**2 + Cy**2

    denom = 2*(Ax*(By - Cy) + Bx*(Cy - Ay) + Cx*(Ay - By))

    ox = (A2*(By - Cy) + B2*(Cy - Ay) + C2*(Ay - By))/denom
    oy = (A2*(Cx - Bx) + B2*(Ax - Cx) + C2*(Bx - Ax))/denom

    # Calculate circumradius:
    # (from https://www.mathalino.com/reviewer/
    # derivation-of-formulas/derivation-of-formula-for-radius-of-circumcircle)
    a = np.sqrt((Ax - Bx)**2 + (Ay-By)**2)
    b = np.sqrt((Bx - Cx)**2 + (By-Cy)**2)
    c = np.sqrt((Cx - Ax)**2 + (Cy-Ay)**2)

    s = (a + b + c) / 2.0
    area = np.sqrt(s * (s - a) * (s - b) * (s - c))

    # circumcircle:
    if area > 0.0:
        rc = a * b * c / (4.0 * area)

    else:
        rc = 0.0

    return ox, oy, rc

def side_check(p):

    # cent = np.mean(p, axis = 0)
    # rads = p - cent
    p1 = np.roll(p, 1, axis=0)

    rads = p1 - p

    Rm =  np.sqrt(rads[:,0]**2 + rads[:,1]**2)
    check_stat = Rm.min()

    return check_stat

def alpha_shape(points, alpha):
    """
    Calculate the alpha_shape of a cluster of points in 2D.

    Parameters
    ----------
    points              A numpy array listing [[x1,y1],[x2,y2]...] for a collection of 2D points

    alpha               The filtering parameter (gauges which triangles to remove from Delaunay triangulation)
                        Note, for this application alpha = 1/d_cell is ideal.

    Returns
    --------
    concave_hull    A list of the indices to vertices in the points structure which define the concave hull
                    (these are all of the points on the boundary).

    Notes
    --------
    Unlike the convex hull, the alpha shape method and concave hull work for complex, concave geometries.
    The result depends on the alpha parameter. A value of alpha = 1/d_cell gives suitable results.

    """

    tri = sps.Delaunay(points)
    tri_edges = []
    # circum_r_list = []

    # loop over triangles:
    # ia, ib, ic = indices of corner points of the
    # triangle
    for ia, ib, ic in tri.vertices:
        pa = points[ia]
        pb = points[ib]
        pc = points[ic]

        # Lengths of sides of triangle
        a = math.sqrt((pa[0]-pb[0])**2 + (pa[1]-pb[1])**2)
        b = math.sqrt((pb[0]-pc[0])**2 + (pb[1]-pc[1])**2)
        c = math.sqrt((pc[0]-pa[0])**2 + (pc[1]-pa[1])**2)

        # Semiperimeter of triangle
        s = (a + b + c)/2.0

        # Area of triangle by Heron's formula
        area = math.sqrt(abs(s*(s-a)*(s-b)*(s-c)))

        if area > 0:
            circum_r = a*b*c/(4.0*area)

        if area == 0:
            circum_r = a*b*c/(4.0*1e-25)

        # Here's the radius filter:
        if circum_r < 1.0/alpha:
            #self.tri_edges_append([ia, ib])
            tri_edges.append([ia, ib])
            tri_edges.append([ib, ic])
            tri_edges.append([ia, ic])

    for i, edge in enumerate(tri_edges):  # First organize the list so that all [i,j] and [j,i] are equalized
        pt1 = edge[0]
        pt2 = edge[1]
        if pt1 > pt2:
            tri_edges[i]=[pt2,pt1]

    tri_edges.sort()

    tri_edges_len = len(tri_edges)
    concave_hull = []

    i = 0
    # Now step through to find and remove all duplicating entities and add them to a new list
    while i < tri_edges_len:
        j = i + 1
        tri_edge_i = tri_edges[i]

        while j < tri_edges_len and tri_edge_i == tri_edges[j]:
            j += 1

        if i == j - 1:
            concave_hull.append(tri_edge_i)
            i += 1
        else:
            i = j

    return concave_hull

def sigmoid(x,g,y_sat):
    """
    A sigmoidal function (logistic curve) allowing user
    to specify a saturation level (y_sat) and growth rate (g).

    Parameters
    ----------
    x            Input values, may be numpy array or float
    g            Growth rate
    y_sat        Level at which growth saturates

    Returns
    --------
    y            Numpy array or float of values

    """
    y = (y_sat*np.exp(g*x))/(y_sat + (np.exp(g*x)-1))

    if np.isnan(y):
        y= 0

    return y

def hill(x,K,n):

    """
    The Hill equation (log-transformed sigmoid). Function ranges
    from y = 0 to +1.

    Parameters
    ----------
    x            Input values, may be numpy array or float. Note all x>0 !
    K            Value of x at which curve is 1/2 maximum (y=0.5)
    n            Hill co-efficient n<1 negative cooperativity, n>1 positive.

    Returns
    --------
    y            Numpy array or float of values

    """
    # assert x.all() > 0

    y = x**n/((K**n)+(x**n))

    return y

def step(t,t_on,t_change):
    """
    A step function (bounded by 0 and 1) based on a logistic curve
    and allowing user to specify time for step to come on (t_on) and time for
    change from zero to one to happen.

    Parameters
    ----------
    t            Input values, may be numpy array or float
    t_on         Time step turns on
    t_change     Time for change from 0 to 1 (off to on)

    Returns
    --------
    y            Numpy array or float of values

    """
    g = (1/t_change)*10
    y = 1/(1 + (np.exp(-g*(t-t_on))))
    # x = g*(t-t_on)
    # y = expit(x)

    # if np.isnan(y):
    #     y= 0

    return y

def pulse(t,t_on,t_off,t_change):
    """
    A pulse function (bounded by 0 and 1) based on logistic curves
    and allowing user to specify time for step to come on (t_on) and time for
    change from zero to one to happen, and time for step to come off (t_change).

    Parameters
    ----------
    t            Input values, may be numpy array or float
    t_on         Time step turns on
    t_off        Time step turns off
    t_change     Time for change from 0 to 1 (off to on)

    Returns
    --------
    y            Numpy array or float of values
    """

    g = (1/t_change)*10
    # x1 = g*(t-t_on)
    # x2 = g*(t-t_off)
    y1 = 1/(1 + (np.exp(-g*(t-t_on))))
    y2 = 1/(1 + (np.exp(-g*(t-t_off))))
    # y1 = expit(x1)
    # y2 = expit(x2)

    y = y1 - y2

    # if np.isnan(y):
    #     y=0

    return y

def H(x):

    y = 0.5*(np.sign(x) +1)

    return y

def emptyDict(dic):
    """
    Tests a dictionary to see if all keys are zero
    and returns True if so.

    Parameters:
    --------------
    dic            a dictionary

    :return:
    """
    # search_success = False
    #
    # while search_success is False:
    for entry in dic.values():

        if entry == 0:
            zero_dic = True

        elif type(entry) == list or type(entry)==str or entry != 0:
            zero_dic = False
            break


    return zero_dic

def grid_vector_data(xpts,ypts,zdata_x,zdata_y,cells,p):

    """
    Takes irregularly spaced vector data in the form of linear arrays of x,y,ux,uy and
    returns grids of X, Y, and Z_x and Z_y. Suitable for streamline plots (streamplot)

    Parameters
    ------------
    xpts                        Linear array of x-coordinates
    ypts                        Linear array of y-coordinates
    zdata_x                     Linear array of vector x-components
    zdata_y                     Linear array of vector y-components
    cells                       Instance of a Cells object

    Returns
    ---------
    X, Y, zi_x, zi_y            Arrays corresponding to xpts, ypts and the vector data points
    """
    # x_full = np.linspace(cells.xmin,cells.xmax,cells.msize)
    # y_full = np.linspace(cells.ymin,cells.ymax,cells.msize)

    xgrid = np.linspace(cells.xmin,cells.xmax,p.isamples)
    ygrid = np.linspace(cells.ymin,cells.ymax,p.isamples)
    X, Y = np.meshgrid(xgrid,ygrid)
    # xgrid = cells.x_v
    # ygrid = cells.y_v
    #
    # X = cells.x_2d
    # Y = cells.y_2d

    # create an interpolation function to resample the cluster mask matrix:
    # mask_funk = interp.interp2d(x_full,y_full,cells.cluster_mask)
    # mask_funk = interp.RectBivariateSpline(x_full,y_full,cells.cluster_mask)
    #
    # new_mask = mask_funk.ev(xgrid,ygrid)

    zi_x = interp.griddata((xpts,ypts),zdata_x,(X,Y))
    zi_x = np.nan_to_num(zi_x)
    # zi_x = np.multiply(zi_x,new_mask)

    zi_y = interp.griddata((xpts,ypts),zdata_y,(X,Y))
    zi_y = np.nan_to_num(zi_y)
    # zi_y = np.multiply(zi_y,new_mask)

    return X,Y,zi_x,zi_y

def griddata(xpts,ypts,zdata,gridsize,cells):

    """
    Takes irregularly spaced data in the form of linear arrays of x,y,z and
    returns grids of X, Y, and Z.

    Parameters
    ------------
    xpts                        Linear array of x-coordinates
    ypts                        Linear array of y-coordinates
    zdata                       Linear array of data values at x-y coordinate points
    gridsize                    Resolution of the interpolation grid (recommended 100x100)

    Returns
    ---------
    X, Y, zi_m                  Arrays corresponding to xpts, ypts and the zdata values. Note
                                zi_m is a masked array.

    """

    # x_full = np.linspace(cells.xmin,cells.xmax,cells.msize)
    # y_full = np.linspace(cells.ymin,cells.ymax,cells.msize)

    xlin = np.linspace(cells.xmin,cells.xmax,gridsize)
    ylin = np.linspace(cells.ymin,cells.ymax,gridsize)

    # create an interpolation function to resample the cluster mask matrix:
    # mask_funk =  interp.RectBivariateSpline(x_full,y_full,cells.cluster_mask)
    # new_mask = mask_funk.ev(xlin,ylin)

    # xmin = np.min(xpts)
    # xmax = np.max(xpts)
    # ymin = np.min(ypts)
    # ymax = np.max(ypts)
    #
    # xlin = np.linspace(xmin,xmax,gridsize)
    # ylin = np.linspace(ymin,ymax,gridsize)
    #
    X,Y = np.meshgrid(xlin,ylin)
#
    zi = interp.griddata((xpts,ypts),zdata,(X,Y))
    zi = np.nan_to_num(zi)
    # zi = np.multiply(new_mask,zi)

    return X, Y, zi

def makegrid(xpts,ypts,gridsize,cells):

    """
    Takes irregularly spaced data in the form of linear arrays of x,y and
    returns grids of X, Y.

    Parameters
    ------------
    xpts                        Linear array of x-coordinates
    ypts                        Linear array of y-coordinates
    gridsize                    Resolution of the interpolation grid (recommended 100x100)

    Returns
    ---------
    X, Y                   Arrays corresponding to xpts, ypts.
    dx, dY                 Array spacing vectors

    """

    # xmin = np.min(xpts)
    # xmax = np.max(xpts)
    # ymin = np.min(ypts)
    # ymax = np.max(ypts)

    # xlin = np.linspace(xmin,xmax,gridsize)
    # ylin = np.linspace(ymin,ymax,gridsize)

    xlin = np.linspace(cells.xmin,cells.xmax,gridsize)
    ylin = np.linspace(cells.ymin,cells.ymax,gridsize)

    X,Y = np.meshgrid(xlin,ylin)

    dx = np.gradient(xlin)
    dy = np.gradient(ylin)

    return X, Y, dx, dy


def RK4(f):
    """
    RK4 (i.e., Runge-Kutta), implemented in the maximally confusing manner.

    It is modified for a differential equation that does not
    depend on time in its rate equation.

    As an example, suppose there is a differential equation
    dc/dt = sqrt(c)

    To solve using RK4, we would define:

    dc = RK4(lambda c: sqrt(c))

    And update it as:
    c1 = co + dc(co,dt)

    Where co is the original value, dt is the time-step
    """

    return lambda y, dt: (
        lambda dy1: (
            lambda dy2: (
                lambda dy3: (
                    lambda dy4: (dy1 + 2 * dy2 + 2 * dy3 + dy4) / 6
                )(dt * f(y + dy3))
            )(dt * f(y + dy2 / 2))
        )(dt * f(y + dy1 / 2))
    )(dt * f(y))


def clip_vals(F, max_value):

    """
    For an array, F, this ensures that min and max values are bounded by
    the value +/- max_val.

    """

    inds_over = (F > max_value).nonzero()
    inds_under = (F < -max_value).nonzero()

    F[inds_over] = max_value
    F[inds_under] = -max_value

    return F



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **export pipeline container** (i.e., container transparently
aggregating runners defined by all available export pipelines) facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseSimPipeRunnerUnsatisfiedException
from betse.lib.matplotlib import mplfigure
from betse.science.phase.phasecls import SimPhase
from betse.science.pipe.export.pipeexpcsv import SimPipeExportCSVs
from betse.science.pipe.export.pipeexpanim import SimPipeExportAnimCells
from betse.science.pipe.export.plot.pipeexpplotcell import (
    SimPipeExportPlotCell)
from betse.science.pipe.export.plot.pipeexpplotcells import (
    SimPipeExportPlotCells)
from betse.util.io.log import logs
from betse.util.type.types import type_check, IterableTypes

# ....................{ CONSTANTS                         }....................
_PIPES_EXPORT_TYPE = (
    SimPipeExportCSVs,
    SimPipeExportPlotCell,
    SimPipeExportPlotCells,
    SimPipeExportAnimCells,
)
'''
Sequence of all export pipeline subclasses in increasing order of expected
duration (i.e., from fastest to slowest), trivially improving aesthetic
responsiveness for end users:

* CSV files, typically exported faster than visuals.
* Single-cell plots, typically exported faster than cell cluster plots.
* Cell cluster plots, typically exported faster than animations.
* Animations.
'''

# ....................{ CLASSES                           }....................
class SimPipesExport(object):
    '''
    High-level **simulation export pipeline container** (i.e., container
    transparently aggregating each pipeline runner defined by each available
    simulation export pipeline).

    Attributes
    ----------
    _PIPES_EXPORT: IterableTypes
        Iterable of all available simulation export pipelines.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self) -> None:
        '''
        Initialize this simulation export pipeline container.
        '''

        # Iterable of all available export pipelines. To avoid dynamically (and
        # inefficiently) regenerating this iterable on each access, this
        # iterable is defined as a tuple rather than generator comprehension.
        self._PIPES_EXPORT = tuple(
            # Export pipeline of this subclass.
            pipe_export_type()
            # For each export pipeline subclass...
            for pipe_export_type in _PIPES_EXPORT_TYPE
        )

    # ..................{ PROPERTIES                        }..................
    # Read-only properties, preventing callers from resetting these attributes.

    @property
    def PIPES_EXPORT(self) -> IterableTypes:
        '''
        Iterable of all available simulation export pipelines.
        '''

        return self._PIPES_EXPORT

    # ..................{ EXPORTERS                         }..................
    @type_check
    def export(self, phase: SimPhase) -> None:
        '''
        Export (e.g., interactively display, non-interactively save) all
        exports enabled for all export pipelines enabled by the passed
        simulation phase.

        Specifically, for each such pipeline:

        * If that pipeline is enabled (i.e., the :meth:`SimPipeABC.is_enabled`
          method for that pipeline returns ``True`` when passed that phase):

          * For each **pipeline runner subconfiguration** (i.e.,
            :class:`SimConfExportABC` instance in the sequence of these
            instances listed by the :meth:`iter_runners_enabled` method for
            that pipeline):

            * Call the method defined by that pipeline implementing this
              runner, passed both this phase and configuration.
            * If that method reports this runner's requirements to be
              unsatisfied (e.g., due to the current simulation configuration
              disabling extracellular spaces), this runner is ignored with a
              non-fatal warning.

        * Else, log an informative message and ignore that pipeline.

        Parameters
        ----------
        phase: SimPhase
            Current simulation phase.
        '''

        # List of 2-tuples "(runner_method, runner_conf)" yielding the method
        # and configuration of each enabled pipeline runner for each export
        # pipeline for this phase, aggregating the 2-tuples yielded by each
        # SimPipeABC.iter_runners_enabled() generator.
        runners_enabled = []

        # For each available export pipeline...
        for pipe_export in self._PIPES_EXPORT:
            # Initialize this pipeline for this phase.
            pipe_export.init(phase)

            # Append all pipeline runners enabled for this pipeline and phase.
            runners_enabled.extend(pipe_export.iter_runners_enabled(phase))

        # Notify the caller of the range of work performed by this subcommand.
        # namely, notify the caller of the total number of times that this
        # method calls the SimCallbacksBC.progressed() callback or a callback
        # calling that callback (e.g., SimCallbacksBC.progressed_next()).
        phase.callbacks.progress_ranged(progress_max=len(runners_enabled))

        # For the method and configuration of each enabled pipeline runner...
        for runner_method, runner_conf in runners_enabled:
            # Metadata associated with this runner.
            runner_metadata = runner_method.metadata

            # Attempt to...
            try:
                # Run this runner with this phase and configuration.
                runner_method(phase, runner_conf)

                #FIXME: Refactor this low-level kludge from the BETSE codebase
                #into a high-level implementation in the BETSEE codebase. See
                #the prominent "FIXME" comment in the "pipeabc" submodule for
                #preliminary work required to begin doing so. For now, this
                #tragically suffices.

                # Notify the caller of the successful completion of this
                # runner. Since the prior call failed to raise an exception,
                # this runner necessarily succeeded.
                phase.callbacks.progressed_next(
                    status='Exported {} "{}".'.format(
                        runner_metadata.noun_singular_lowercase,
                        runner_metadata.kind))
            # If this runner's requirements are unsatisfied (e.g., due to the
            # current simulation configuration disabling fluid flow), notify
            # the caller of this non-fatal condition and continue.
            except BetseSimPipeRunnerUnsatisfiedException as exception:
                phase.callbacks.progressed_next(
                    status='Excluding {} "{}", as {}.'.format(
                        runner_metadata.noun_singular_lowercase,
                        runner_metadata.kind,
                        exception.reason))
            # Else if this runner raises any other exception, permit this
            # exception to propagate up the callstack without intervention.

        # Unconditionally close all currently open matplotlib figures
        # regardless of whether any of the above runners invoked matplotlib.
        #
        # In theory, each of the runner_method() methods called above should
        # explicitly do so already; in practice, the matplotlib API is
        # sufficiently non-deterministic *AND* resource-consumptive to justify
        # such precautions.
        mplfigure.close_figures_all()

        # Log the directory to which all results were exported.
        logs.log_info('Simulation results exported to:')
        logs.log_info('\t%s', phase.export_dirname)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
CLI-specific functional tests exercising all simulation subcommands pertaining
to gene regulatory networks (e.g., `betse sim-grn`, `betse plot sim-grn`).
'''

# ....................{ IMPORTS                            }....................
import pytest
from betse.util.test.pytest.mark.pytskip import skip_unless_lib_runtime_optional

# ....................{ DECORATORS                         }....................
skip_unless_networkable = skip_unless_lib_runtime_optional('networkx', 'pydot')
'''
Decorator skipping the decorated test if either of these optional runtime
dependencies are unavailable, both of which are required by network plotting
subcommands (e.g., ``plot sim-grn``).
'''

# ....................{ TESTS                              }....................
@skip_unless_networkable
@pytest.mark.parametrize(
    ('unpickle_phase_name', 'init_subcommands'), (
        ('seed', (('seed',),),),
        ('init', (('seed',), ('init',),),),
        ('sim',  (('seed',), ('init',), ('sim',),),),
    ))
def test_cli_grn_isolated(
    betse_cli_sim: 'CLISimTester',
    unpickle_phase_name: str,
    init_subcommands: tuple,
) -> None:
    '''
    Test simulating the default gene regulatory network (GRN) isolated away from
    all bioelectrical phenomena for the passed unpickle simulation phase and
    simulation subcommands.

    Parameters
    ----------
    betse_cli_sim : CLISimTester
        Object running BETSE CLI simulation subcommands.
    unpickle_phase_name : str
        Lowercase name of the type of **unpickle simulation phase** (i.e.,
        previously pickled simulation phase to unpickle as the computational
        basis for the current network to be run by the ``betse sim-grn``
        subcommand) to be exercised by this test.
    init_subcommands : tuple
        Tuple of all simulation subcommands to be exercised by this test
        *before* exercising the ``betse sim-grn`` subcommand.
    '''

    # Defer heavyweight imports required for logging. For readability in the
    # unfortunate event of exceptions raised by subsequent imports, defer all
    # remaining imports.
    from betse.util.io.log import logs

    # Log this networking attempt with suitable aesthetics.
    logs.log_banner(
        title='sim-grn ({})'.format(unpickle_phase_name), padding='~')

    # Defer all remaining heavyweight imports.
    from betse.science.enum.enumconf import GrnUnpicklePhaseType
    from betse.util.path import pathnames
    from betse.util.type import enums

    # Simulation configuration specific to this test.
    p = betse_cli_sim.sim_state.p

    # Enable the saving of visuals, preventing the "plot sim-grn" subcommand
    # tested below from silently reducing to a noop.
    betse_cli_sim.sim_state.config.enable_visuals_save()

    #FIXME: Replace this overkill method call with direct usage of the local
    #"p" parameter; then remove this method entirely from the codebase.
    # Enable networking.
    betse_cli_sim.sim_state.config.enable_networks()

    # Enable this type of unpickle simulation phase by mapping from the passed
    # string to the corresponding enumeration member. Ideally, this member
    # rather than this string would simply be passed to this test;
    # unfortunately, doing so would require unsafely importing from the main
    # codebase at module scope, encouraging non-human-readable exceptions.
    p.grn_unpickle_phase_type = enums.get_member_from_name_uppercased(
        enum_type=GrnUnpicklePhaseType, enum_member_name=unpickle_phase_name)

    # Test all simulation subcommands required by this GRN-specific subcommand
    # with this configuration for this phase type.
    betse_cli_sim.run_subcommands(*init_subcommands)

    # Test this GRN-specific subcommand from scratch.
    betse_cli_sim.run_subcommands(('sim-grn',),)

    # Log this re-networking attempt with suitable aesthetics.
    logs.log_banner(
        title='sim-grn (re-{})'.format(unpickle_phase_name), padding='~')

    # Prepare to rerun the "sim-grn" subcommand from the prior run pickled by
    # the prior subcommand. To do so safely (in order):
    #
    # * Reconstruct the relative filename of the prior GRN run.
    # * Ensure that the next GRN run is pickled to another file.
    p.grn_unpickle_filename_relative = pathnames.join(
        p.grn_pickle_dirname_relative, p.grn_pickle_basename)
    p.grn_pickle_basename = 'new_' + p.grn_pickle_basename

    # Redefine all absolute pathnames depending upon these relative pathnames.
    p.reload_paths()

    # Test both this GRN-specific subcommand from the prior such run and, for
    # completeness, exporting the results of doing so.
    betse_cli_sim.run_subcommands(('sim-grn',), ('plot', 'sim-grn',),)


@skip_unless_networkable
def test_cli_sim_grn_integrated(betse_cli_sim: 'CLISimTester') -> None:
    '''
    Test simulating the default gene regulatory network (GRN) integrated
    together with all bioelectrical phenomena.

    Parameters
    ----------
    betse_cli_sim : CLISimTester
        Object running BETSE CLI simulation subcommands.
    '''

    # Enable these networks.
    betse_cli_sim.sim_state.config.enable_networks()

    # Test all default simulation-specific subcommands with this configuration.
    betse_cli_sim.run_subcommands_try()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **frozen application** (i.e., cross-platform Python application
converted into one or more platform-specific executable binaries) facilities.
'''

# ....................{ IMPORTS                            }....................
import sys
from betse.exceptions import BetsePyFrozenException
# from betse.util.type.types import type_check, SequenceTypes

# ....................{ TESTERS                            }....................
def is_frozen() -> bool:
    '''
    ``True`` only if the active Python interpreter is **frozen** (i.e., embedded
    in a platform-specific compressed executable archiving this application and
    all transitive dependencies thereof).
    '''

    # If the "sys" module has an attribute:
    #
    # * "_MEIPASS", this is a binary frozen by PyInstaller.
    # * "frozen", this is a binary frozen by a non-PyInstaller freezer (e.g.,
    #   "py2app", "py2exe").
    return is_frozen_pyinstaller() or hasattr(sys, 'frozen')


def is_frozen_pyinstaller() -> bool:
    '''
    ``True`` only if the active Python interpreter is frozen with PyInstaller.

    This function returns ``True`` only if the PyInstaller-specific private
    attribute ``_MEIPASS`` added to the canonical :mod:`sys` module by the
    PyInstaller bootloader embedded in this frozen executable (if any) exists.
    '''

    # Hear no evil, code no evil, comment no evil.
    return hasattr(sys, '_MEIPASS')

# ....................{ GETTERS                            }....................
def get_app_dirname_pyinstaller() -> str:
    '''
    Absolute path of the temporary directory extracted by the PyInstaller
    bootloader from the platform-specific executable binary frozen by
    PyInstaller for this this application.

    This directory contains all files and directories required to run this
    application, including both Python modules, packages, and C extensions _and_
    non-Python resources.

    Returns
    ----------
    str
        Absolute path of this directory.

    Raises
    ----------
    :exc:`betse.exceptions.BetsePyFrozenException`
        If this application is _not_ frozen with PyInstaller.
    '''

    # If this application is *NOT* frozen with PyInstaller, raise an exception.
    if not is_frozen_pyinstaller():
        raise BetsePyFrozenException('Application not frozen with PyInstaller.')

    # That is not buggy which can eternal lie.
    return sys._MEIPASS



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Voltage-gated sodium channel classes.
'''

# .................... IMPORTS                            ....................
from abc import ABCMeta, abstractmethod
import numpy as np
from betse.science import sim_toolbox as stb
from betse.science.channels.channelsabc import ChannelsABC
from betse.science.math import toolbox as tb
from betse.util.io.log import logs

# .................... BASE                               ....................
class VgNaABC(ChannelsABC, metaclass=ABCMeta):
    '''
    Abstract base class of all Voltage-gated sodium channel classes.

    Attributes
    ----------
    _mInf :
        Equillibrium value function for m-gates of channel
    _mTau :
        Time-constant function for m-gates of channel
    _hInf :
        Equillibrium value function for h-gates of channel
    _hTau :
        Time-constant function for h-gates of channel
    '''

    def init(self, vm, cells, p, targets = None):
        '''
        Initialize targeted voltage-gated sodium channels at the initial
        time step of the simulation based on the initial cell Vmems.

        Channel model uses Hodgkin-Huxley kinetic model style
        for voltage gated channels.
        '''

        if targets is None:

            self.targets = None

            if type(vm) == float or type(vm) == np.float64:
                self.data_length = 1

            else:
                self.data_length = len(vm)


        else:
            self.targets = targets
            self.data_length = len(self.targets)
            self.mdl = len(cells.mem_i)

        self.modulator = 1.0

        self.v_corr = 0.0  # in experiments, the measurement junction voltage is about 10 mV

        if self.targets is None:

            V = vm * 1000 + self.v_corr

        else:
            V = vm[self.targets] * 1000 + self.v_corr

        self._init_state(V)

        self.ions = ['Na']
        self.rel_perm = [1.0]

    def run(self, vm, p):
        '''
        Handle all targeted voltage-gated sodium channels by working with the passed
        user-specified parameters on the tissue simulation and cellular
        world for a time step.

        Channel model uses Hodgkin-Huxley kinetic model style
        for voltage gated channels.

        '''

        if self.targets is None:

            V = vm*1000 + self.v_corr

        else:

            V = vm[self.targets] * 1000 + self.v_corr

        self._calculate_state(V)

        self._implement_state(V, p)

    def _implement_state(self, V, p):

        # Update the 'm' and 'h' gating functions:
        self.update_mh(p, time_unit = self.time_unit)

        # calculate the open-probability of the channel:
        P = (self.m ** self._mpower) * (self.h ** self._hpower)

        if self.targets is None:

            self.P = P

        else:
            self.P = np.zeros(self.mdl)
            self.P[self.targets] = P


    @abstractmethod
    def _init_state(self, V):
        '''
        Do something.
        '''
        pass


    @abstractmethod
    def _calculate_state(self, V):
        '''
        Do something.
        '''
        pass

# ....................{ SUBCLASS                           }....................
class Nav1p2(VgNaABC):
    '''
    NaV model from Hammil et al 1991, derived from rat neocortical neurons.

    This channel produces well behaved action-potentials with a variety of vgK channels. Good general-purpose
    vgNa channel.

    Reference: Hammil, OP et al. Patch-clamp studies of voltage-gated currents in identified neurons
    of the rat cerebral cortex. Cereb. Cortex, 1991, 1, 48-61

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgNa channel type: Nav1.2')

        # time scale of the model (most models are in miliseconds, some are in seconds)
        self.time_unit = 1.0e3

        self.vrev = 50     # reversal voltage used in model [mV]

        mAlpha = (0.182 * ((V - 10.0) - -35.0)) / (1 - (np.exp(-((V - 10.0) - -35.0) / 9)))
        mBeta = (0.124 * (-(V - 10.0) - 35.0)) / (1 - (np.exp(-(-(V - 10.0) - 35.0) / 9)))

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = mAlpha / (mAlpha + mBeta)
        self.h = 1.0 / (1 + np.exp((V - -65.0 - 10.0) / 6.2))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 3
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        mAlpha = (0.182 * ((V - 10.0) - -35.0)) / (1 - (np.exp(-((V - 10.0) - -35.0) / 9)))
        mBeta = (0.124 * (-(V - 10.0) - 35.0)) / (1 - (np.exp(-(-(V - 10.0) - 35.0) / 9)))

        self._mInf = mAlpha / (mAlpha + mBeta)
        self._mTau = (1 / (mAlpha + mBeta))
        self._hInf = 1.0 / (1 + np.exp((V - -65.0 - 10.0) / 6.2))
        self._hTau = (1 / (
            (0.024 * ((V - 10.0) - -50.0)) /
            (1 - (np.exp(-((V - 10.0) - -50.0) / 5))) + (
                0.0091 * (-(V - 10.0) - 75.000123)) / (1 - (np.exp(-(-(V - 10) - 75.000123) / 5)))))

class Nav1p3(VgNaABC):

    """
    Nav1.3 sodium channel TTX sensitive expressed in embryonic and neonatal tissue, rare in
    adults.

    It is thought that the fast activation and inactivation kinetics of Nav1.3, together with its rapid repriming
    kinetics and persistent current component, contributes to the development of spontaneous ectopic
    discharges and sustained rates of firing characteristics of injured sensory nerves

    Nav1.3 is predominantly expressed during pre-myelination stages of development in somatodendritic and axonal
    compartments of embryonic neurons. In rodents, the expression is attenuated after birth while in humans,
    the channel is highly expressed in somatodendritic compartment of myelinated neurons. Axotomy or other forms of
    nerve damage lead to the reexpression of NaV1.3 and the associated beta-3 subunit in sensory neurons,
    but not in primary motor neurons

    reference: Cummins et al. Nav1.3 sodium channel: Rapid repriming and slow closed-state
    inactivation display quantitative differences after expression in a mammalian
    cell line and in spinal sensory neurons. J. Neurosci., 2001, 21, 5952-61

    """

    def _init_state(self, V):

        logs.log_info('You are using the vgNa channel: Nav1p3')

        # time scale of the model (most models are in miliseconds, some are in seconds)
        self.time_unit = 1.0e3

        self.vrev = 50     # reversal voltage used in model [mV]

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 3
        self._hpower = 1

        mAlpha = (0.182 * ((V) - -26)) / (1 - (np.exp(-((V) - -26) / 9)))

        mBeta = (0.124 * (-(V) - 26)) / (1 - (np.exp(-(-(V) - 26) / 9)))

        self.m = mAlpha / (mAlpha + mBeta)
        self.h = 1 / (1 + np.exp((V - (-65.0)) / 8.1))

    def _calculate_state(self, V):


        mAlpha = (0.182 * ((V) - -26)) / (1 - (np.exp(-((V) - -26) / 9)))

        mBeta = (0.124 * (-(V) - 26)) / (1 - (np.exp(-(-(V) - 26) / 9)))

        self._mInf = mAlpha / (mAlpha + mBeta)
        self._mTau = 1 / (mAlpha + mBeta)
        self._hInf = 1 / (1 + np.exp((V - (-65.0)) / 8.1))
        self._hTau = 0.40 + (0.265 * np.exp(-V / 9.47))

class NavRat2(VgNaABC):

    """
    Generic vgNa channel, slow to inactivate.

    Reference:  	McCormick DA. et al. A model of the electrophysiological properties of
    thalamocortical relay neurons. J. Neurophysiol., 1992 Oct , 68 (1384-400).
    """

    def _init_state(self, V):

        logs.log_info('You are using the vgNa channel: NavRat2')

        self.time_unit = 1.0e3

        self.vrev = 50  # reversal voltage used in model [mV]

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 3
        self._hpower = 1

        mAlpha = 0.091 * (V + 38) / (1 - np.exp((-V - 38) / 5))
        mBeta = -0.062 * (V + 38) / (1 - np.exp((V + 38) / 5))

        self.m = mAlpha / (mAlpha + mBeta)

        hAlpha = 0.016 * np.exp((-55 - V) / 15)
        hBeta = 2.07 / (np.exp((17 - V) / 21) + 1)

        self.h = hAlpha / (hAlpha + hBeta)


    def _calculate_state(self, V):

        mAlpha = 0.091 * (V + 38) / (1 - np.exp((-V - 38) / 5))
        mBeta = -0.062 * (V + 38) / (1 - np.exp((V + 38) / 5))

        self._mInf = mAlpha / (mAlpha + mBeta)
        self._mTau = 1 / (mAlpha + mBeta)

        hAlpha = 0.016 * np.exp((-55 - V) / 15)
        hBeta = 2.07 / (np.exp((17 - V) / 21) + 1)

        self._hInf = hAlpha / (hAlpha + hBeta)
        self._hTau = 1 / (hAlpha + hBeta)

class NavRat1(VgNaABC):

    """
    Generic vgNa channel, good activity.

    Reference: Huguenard JR. et al. Developmental changes in Na+ conductances in rat neocortical neurons:
     appearance of a slowly inactivating component. J. Neurophysiol., 1988 Mar , 59 (778-95).
    """

    def _init_state(self, V):

        logs.log_info('You are using the vgNa channel: NavRat1')

        # time scale of the model (most models are in miliseconds, some are in seconds)
        self.time_unit = 1.0e3

        self.vrev = 50  # reversal voltage used in model [mV]

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 3
        self._hpower = 1

        mAlpha = (0.182 * (V - -35)) / (1 - (np.exp(-(V - -35) / 9)))
        mBeta = (0.124 * (-V - 35)) / (1 - (np.exp(-(-V - 35) / 9)))

        self.m = mAlpha / (mAlpha + mBeta)
        self.h = 1.0 / (1 + np.exp((V - -65) / 6.2))


    def _calculate_state(self, V):


        mAlpha = (0.182 * (V - -35)) / (1 - (np.exp(-(V - -35) / 9)))
        mBeta = (0.124 * (-V - 35)) / (1 - (np.exp(-(-V - 35) / 9)))

        self._mInf = mAlpha / (mAlpha + mBeta)
        self._mTau = 1 / (mAlpha + mBeta)

        self._hInf = 1.0 / (1 + np.exp((V - -65) / 6.2))
        self._hTau = 1 / ((0.024 * (V - -50)) / (1 - (np.exp(-(V - -50) / 5))) + (0.0091 * (-V - 75.000123)) / (
        1 - (np.exp(-(-V - 75.000123) / 5))))

class NavRat3(VgNaABC):

    """
    Generic vgNa channel.

    Reference:  	McCormick DA. et al. A model of the electrophysiological properties of
    thalamocortical relay neurons. J. Neurophysiol., 1992 Oct , 68 (1384-400).
    """

    def _init_state(self, V):

        logs.log_info('You are using the vgNa channel: NavRat3')

        # time scale of the model (most models are in miliseconds, some are in seconds)
        self.time_unit = 1.0e3

        self.vrev = 50  # reversal voltage used in model [mV]

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 3
        self._hpower = 1

        mAlpha = (0.182 * (V - -35)) / (1 - (np.exp(-(V - -35) / 9)))
        mBeta = (0.124 * (-V - 35)) / (1 - (np.exp(-(-V - 35) / 9)))

        self.m = mAlpha / (mAlpha + mBeta)
        self.h = 1.0 / (1 + np.exp((V - -65) / 6.2))


    def _calculate_state(self, V):


        mAlpha = (0.182 * (V - -35)) / (1 - (np.exp(-(V - -35) / 9)))
        mBeta = (0.124 * (-V - 35)) / (1 - (np.exp(-(-V - 35) / 9)))

        self._mInf = mAlpha / (mAlpha + mBeta)
        self._mTau = 1 / (mAlpha + mBeta)

        self._hInf = 1.0 / (1 + np.exp((V - -65) / 6.2))
        self._hTau = 1 / ((0.024 * (V - -50)) / (1 - (np.exp(-(V - -50) / 5))) + (0.0091 * (-V - 75.000123)) / (
        1 - (np.exp(-(-V - 75.000123) / 5))))

class NaLeak(VgNaABC):

    '''
    Simple sodium leak channel -- always open -- for substance modulation.

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using an Na+ leak channel')

        # time scale of the model (most models are in miliseconds, some are in seconds)
        self.time_unit = 1.0

        self.vrev = 50     # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = np.ones(self.data_length)
        self.h = np.ones(self.data_length)

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 0
        self._hpower = 0


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1
        self._mTau = 1
        self._hInf = 1
        self._hTau = 1

class Nav1p6(VgNaABC):

    """
    Nav1.6 was detected during the embryonic period in brain. This is the most abundantly expressed
    NaV channel in the CNS during adulthood. Is the most abundant channel at mature nodes of Ranvier
    in myelinated axons in the CNS. NaV1.6 is broadly expressed in the nervous system in a variety
    of cells including Purkinje cells, motor neurons, pyramidal and granule neurons, glial cells and
    Schwann cells and is enriched at the nodes of Ranvier.
    Nav1.6 channels have been also detected in immune cells, such microglia and macrophagues
     and in cultured microglia, Nav1.6 is the most prominently expressed sodium channel.

     Reference: Smith MR. et al. Functional analysis of the mouse Scn8a sodium channel. \
     J. Neurosci., 1998 Aug 15 , 18 (6093-102).

    """

    def _init_state(self, V):

        logs.log_info('You are using the persistent vgNa channel: Nav1p6')

        # time scale of the model (most models are in miliseconds, some are in seconds)
        self.time_unit = 1.0e3

        self.vrev = 50     # reversal voltage used in model [mV]

        self._mpower = 1.0
        self._hpower = 0.0

        self.m = 1.0000 / (1 + np.exp(-0.03937 * 4.2 * (V - -17.000)))
        self.h = 1.0

    def _calculate_state(self, V):

        self._mInf = 1.0000 / (1 + np.exp(-0.03937 * 4.2 * (V - -17.000)))
        self._mTau = 1

        self._hInf = 1.0
        self._hTau = 1.0

# class Nav(VgNaABC):
#
#     """
#     Minimal, general model of voltage gated sodium channels, from Pospischil et al 2008.
#
#
#      Reference: Pospischil M. et al. Minimal Hodgkin–Huxley type models for different classes
#      of cortical and thalamic neurons. Biological Cybernetics., 2008.
#
#      This model was not obtained from Channelpedia.
#
#     """
#
#     def _init_state(self, V, sim, p):
#
#         logs.log_info('You are using the general Nav channel: Nav')
#
#         self.time_unit = 1.0
#
#         self.vrev = 50     # reversal voltage used in model [mV]
#
#         self._mpower = 3.0
#         self._hpower = 1.0
#
#
#         self._alpha_m = -(0.32*(V - 13))/(np.exp(-(V-13)/4)-1)
#         self._beta_m = -(0.28 * (V - 40)) / (np.exp((V - 40) / 5) - 1)
#
#         self._alpha_h = 0.128*np.exp((V - 17)/18)
#         self._beta_h = 4/ (np.exp(-(V - 40) / 5) + 1)
#
#         self.m = self._alpha_m/(self._alpha_m + self._beta_m)
#         self.h = self._alpha_h/(self._alpha_h + self._beta_h)
#
#     def _calculate_state(self, V, sim, p):
#
#         self._alpha_m = -(0.32*(V - 13))/(np.exp(-(V-13)/4)-1)
#         self._beta_m = -(0.28 * (V - 40)) / (np.exp((V - 40) / 5) - 1)
#
#         self._alpha_h = 0.128*np.exp((V - 17)/18)
#         self._beta_h = 4/ (np.exp(-(V - 40) / 5) + 1)














#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level command-line argument facilities.
'''

#FIXME: It'd be great to contribute this back to the official "argparse" module
#by contributing a patch. See:
#    http://bugs.python.org/issue12806

# ....................{ IMPORTS                           }....................
import re
from argparse import HelpFormatter
# from betse.util.io.log import logs
from betse.util.type.types import type_check, GeneratorType, SequenceTypes

# ....................{ CONSTANTS                         }....................
_BLANK_LINE_REGEX = re.compile(r'^\s*$')
'''
Regular expression matching only blank lines.
'''


_INDENTATION_REGEX = re.compile(r'^( *)')
'''
Regular expression capturing all spaces prefixing the subject string.
'''

# ....................{ FORMATTERS                        }....................
class SemicolonAwareHelpFormatter(HelpFormatter):
    '''
    Formatter wrapping all lines *not* prefixed by ``;`` in
    :class:`ArgumentParser` help text into paragraphs.

    Newlines in all lines prefixed by ``;`` will be preserved as is, preventing
    such lines from being wrapped. Newlines in all other lines will be
    implicitly converted to single spaces. (See examples below.)

    See Also
    ----------
    http://bugs.python.org/issue12806
    http://bugs.python.org/file28091/paraformatter.py
        argparse's lack of official paragraph formatter is long recognized, as
        documented by the above bug thread. A year and a half after the first
        post in such thread, rurpy2 contributed a module "paraformatter", which
        this module remains strongly inspired by.

    Examples
    ----------
        >>> from argparse import ArgumentParser
        >>> from betse.util.os.command import HelpFormatterParagraph
        >>> arg_parser = ArgumentParser(
        ...     formatter_class = HelpFormatterParagraph)
        >>> arg_subparsers = arg_parser.add_subparsers()
        >>> arg_subparsers.add_parser(
        ...     name = 'road',
        ...     description = """
        ...         The Road goes ever on and on
        ...         Down from the door where it began.
        ...         Now far ahead the Road has gone,
        ...         ;And I must follow, if I can,
        ...         ;Pursuing it with eager feet,
        ...         ;Until it joins some larger way,
        ...         ;Where many paths and errands meet.
        ...
        ...            The Road goes ever on and on
        ...            Down from the door where it began.
        ...            Now far ahead the Road has gone,
        ...            And I must follow, if I can,
        ...         ;  Pursuing it with weary feet,
        ...         ;  Until it joins some larger way,
        ...         ;  Where many paths and errands meet.
        ...         ;  And whither then? I cannot say.
        ...     """)
        >>> arg_parser.parse_args('road --help')
        The Road goes ever on and on Down from the door where it began. Now far
        ahead the Road has gone,
        And I must follow, if I can,
        Pursuing it with eager feet,
        Until it joins some larger way,
        Where many paths and errands meet.

        The Road goes ever on and on Down from the door where it began. Now far
        ahead the Road has gone, And I must follow, if I can,
        Pursuing it with weary feet,
        Until it joins some larger way,
        Where many paths and errands meet.
        And whither then? I cannot say.
    '''

    # ..................{ SUPERCLASS                        }..................
    def _fill_text(self, text: str, width: int, indent: str) -> str:
        '''
        Passed string with all non-indented lines wrapped into paragraphs.
        '''

        return '\n'.join(self._split_lines_indented(text, width, indent))


    def _split_lines(self, text: str, width: int) -> list:
        '''
        List of lines produced by wrapping all non-indented lines in the passed
        text into paragraphs and splitting such string on newlines.
        '''

        return self._split_lines_indented(text, width)

    # ..................{ SUBCLASS                          }..................
    @type_check
    def _split_lines_indented(
        self, text: str, width: int, indent: str = '') -> list:
        '''
        List of lines produced by wrapping all non-indented lines in the passed
        text into paragraphs and splitting this string on newlines.
        '''

        # Avoid circular import dependencies.
        from betse.util.type.text.string import strs

        # Initialize this text wrapper with the passed arguments.
        strs.text_wrapper_config.width = width
        strs.text_wrapper_config.initial_indent = indent
        strs.text_wrapper_config.subsequent_indent = indent

        # List of lines wrapped from this text.
        lines = []

        # Iteratively construct this list.
        for paragraph_wrapped_lines in self._paragraphs_wrapped_lines(text):
            lines.extend(paragraph_wrapped_lines)

        # Return this list.
        return lines


    @type_check
    def _paragraphs_wrapped_lines(self, text: str) -> GeneratorType:
        '''
        Generator yielding a list of all lines comprising the next paragraph
        wrapped from the passed text.
        '''

        # Number of spaces indenting the first line of this text.
        main_indent_len = _get_text_indent_len(text)

        # List of lines accumulating the current paragraph. For efficiency,
        # this would ideally be a StringIO-based string buffer rather than a
        # list; unfortunately, the former fails to provide a sane method for
        # testing whether this buffer is empty (i.e., has been written to),
        # whereas the latter implicitly provides such functionality.
        paragraph_lines = []

        # For each line split from this text...
        for line in text.splitlines():
            # If this line is a blank line, pretend this line consisted of a
            # ";" instead to force this line to stop wrapping the currently
            # accumulated paragraph if any.
            if _BLANK_LINE_REGEX.match(line):
                line = ';'
            assert len(line), 'Line empty.'

            # If this line is prefixed by ";", prevent this line from being
            # wrapped by yielding such line as a newly unwrapped paragraph.
            #
            # Since the prior conditional guarantees this line to now be
            # nonempty, the first character of this line is safely indexable.
            if line[0] == ';':
                # print('Non-wrappable line detected: ' + line)

                # Strip this ";".
                line = line[1:]

                # If a previously accumulated paragraph exists, yield this
                # paragraph *BEFORE* yielding this line.
                if paragraph_lines:
                    yield self._wrap_lines(paragraph_lines, main_indent_len)

                    # Clear such paragraph.
                    paragraph_lines = []

                # Yield such line as a discrete paragraph.
                yield [line]
            # Else, wrap this line. Specifically, accumulate this line for
            # subsequent wrapping and yielding.
            else:
                # print('Wrappable line detected: ' + line)
                paragraph_lines.append(line)

        # If a previously accumulated paragraph exists, yield this as the last.
        if paragraph_lines:
            yield self._wrap_lines(paragraph_lines, main_indent_len)


    @type_check
    def _wrap_lines(self, lines: SequenceTypes, main_indent_len: int) -> list:
        '''
        Join the passed list of lines on newline, reduce two or more
        consecutive whitespace characters in the passed string to one space,
        strip leading and trailing whitespace, wrap the result, and return a
        wrapped list of lines.

        Since the base class performs the same reduction and stripping of
        whitespace, so do we.
        '''

        # Avoid circular import dependencies.
        from betse.util.type.text.string import strjoin, strs

        # Difference in spaces between the indentation of the first line of the
        # entire text passed above and the indentation of the first line of the
        # currently passed paragraph.
        paragraph_indent_len = _get_text_indent_len(lines[0]) - main_indent_len

        # String of spaces of this length.
        paragraph_indent = ' ' * paragraph_indent_len

        # Indent each wrapped line by this difference.
        strs.text_wrapper_config.initial_indent    = paragraph_indent
        strs.text_wrapper_config.subsequent_indent = paragraph_indent

        # Strip and wrap these lines.
        return strs.text_wrapper_config.wrap(
            self._whitespace_matcher.sub(
                ' ', strjoin.join_on_newline(lines)).strip())

# ....................{ GETTERS                           }....................
@type_check
def _get_text_indent_len(text: str) -> int:
    '''
    Number of spaces indenting the first line of the passed text.
    '''

    return len(_INDENTATION_REGEX.match(text).group(1))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Test-specific simulation configuration classes wrapping low-level dictionaries
both serialized to and deserialized from on-disk YAML-formatted files.
'''

#FIXME: This submodule requires heavy refactoring away from the current
#low-level approach in favour of the new high-level "confabc"-based approach --
#namely, the "YamlABC" base class coupled with the conf_alias() data
#descriptor. Together, this new functionality provides a substantially better
#YAML-to-Python-object-mapping (YPOM) than the ad-hoc and overly verbose
#boilerplate implemented below. Specifically:
#
#* Eliminate all properties defined below.
#* Replace all usage of the low-level "self.p.conf" dictionary with high-level
#  data descriptors defined by "self.p".
#FIXME: Ideally, after implementing the above, use of the
#"SimConfigTestWrapper" wrapper will be able to be replaced everywhere in tests
#by direct use of the "SimConfTestInternal.p" property providing direct access
#to this "Parameters" object, which increasingly provides all test
#functionality. We're not quite there yet -- but we will be, eventually.

# ....................{ IMPORTS                            }....................
# This subclass imports from submodules defined by the main codebase and is
# thus *NOT* safely importable from fixture submodules directly imported by
# "conftest" plugin modules. To defer the importation of this submodule until
# *AFTER* test collection, this submodule is intentionally segregated.
from betse.science.enum.enumconf import IonProfileType, SolverType
from betse.science.enum.enumphase import SimPhaseKind
from betse.science.parameters import Parameters
from betse.science.phase.phasecls import SimPhase
from betse.science.phase.require import phasereqs
from betse.science.phase.require.abc.phasereqset import (
    SimPhaseRequirementsOrNoneTypes)
from betse.science.pipe.export.pipeexps import SimPipesExport
from betse.util.io.log import logs
from betse.util.type.types import type_check

# ....................{ SUPERCLASSES                       }....................
class SimConfigTestWrapper(object):
    '''
    **Test-specific simulation configuration wrapper** (i.e., object wrapping a
    Python dictionary or sequence deserialized from a YAML-formatted simulation
    configuration file).

    This wrapper is intended to be instantiated *only* by non-interactive
    automation (e.g., tests, scripts).

    This wrapper superficially wraps this Python container with convenience
    methods safely modifying this container. This wrapper is lower level than
    the high-level :class:`Parameters` simulation configuration, which
    transforms this container into numerous high-level objects rather than
    merely wrapping this container. While the :class:`Parameters` configuration
    is principally used by backend simulation modelling in the
    :mod:`betse.science` package, this wrapper is principally used by frontend
    logic modifying simulation configurations on behalf of either interactive
    users (e.g., the BETSEE GUI) or automated tests.

    Attributes
    ----------
    p : Parameters
        High-level simulation configuration encapsulated by this test wrapper.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, p: Parameters) -> None:
        '''
        Wrap the passed simulation configuration.

        Parameters
        ----------
        p : Parameters
            Simulation configuration to be wrapped.
        '''

        # Classify all passed parameters.
        self.p = p

    # ..................{ PROPERTIES ~ float                 }..................
    @property
    def environment_size(self) -> float:
        '''
        Square dimension in meters of the external environment containing the
        cell cluster for this configuration.

        For simplicity, BETSE constrains the environment to be square in shape.
        This dimension thus defines both the environmental width *and* height.
        '''

        # Coerce the current number to a float for safety.
        return float(self.p.conf['world options']['world size'])


    @environment_size.setter
    @type_check
    def environment_size(self, environment_size: float) -> None:
        '''
        Set the square dimension in meters of the external environment
        containing the cell cluster for this configuration.
        '''

        # Coerce the passed number to a float for safety.
        self.p.conf['world options']['world size'] = float(environment_size)

    # ..................{ MINIMIZERS                         }..................
    #FIXME: Additionally, the three time durations defined by the
    #"self.p.conf['gene regulatory network settings']['sim-grn settings']"
    #dictionary must also be minified. For simplicity, unconditionally minify
    #these settings within this method regardless of whether the "sim-grn"
    #subcommand is being exercised by the current test or fixture.
    def minify(self) -> None:
        '''
        Minimize the space and time costs associated with running the
        simulation configured by this configuration while preserving all
        fundamental configuration features.

        Specifically, this method numerically reduces all configuration options
        pertaining to either world size *or* simulation time to their **minimum
        permissible values** (i.e., the smallest values still preserving
        simulation stability). This method is intended to be called only by
        test automation.
        '''

        # Minify initialization time to exactly three sampled time steps. For
        # safety, permit currently defined options smaller than the minimums
        # defined below to override these minimums.

        # Duration of each time step in seconds.
        self.p.init_time_step = min(self.p.init_time_step, 1.0e-3)

        # Interval to sample these steps at in seconds.
        self.p.init_time_sampling = min(
            self.p.init_time_sampling, self.p.init_time_step)

        # Total simulation time in seconds. The first digit effectively defines
        # the number of sampled time steps, by the above choice of time step.
        self.p.init_time_total = self.p.init_time_step * 3

        # Minify simulation time to the same durations. To ensure that
        # property setter validation compares durations in the expected manner,
        # properties are assigned in order of increasing duration.
        self.p.sim_time_step = min(
            self.p.sim_time_step, self.p.init_time_step)
        self.p.sim_time_sampling = min(
            self.p.sim_time_sampling, self.p.sim_time_step)
        self.p.sim_time_total = self.p.sim_time_step * 3

        #FIXME: Generalize to minify the time steps of *ALL* enabled events.
        # Minify simulation event times to the same durations in a manner
        # preserving the relative time that these events occur with respect to
        # the pre-minified simulation duration in seconds. For simplicity, each
        # such minification preserves the following ratio:
        #
        #     event_time_old       event_time_new
        #     ------------------ = ------------------
        #     sim_time_total_old   sim_time_total_new
        #
        # Since "sim_time_total_new" is given by "self.p.sim_time_total", the
        # desired unknown "event_time_new" is given by:
        #
        #                      self.p.sim_time_total * event_time_old
        #                      ---------------------------------------
        #     event_time_new = sim_time_total_old
        #FIXME: Well, we've eliminated the "event_cut['cut time']" key entirely.
        #Nonetheless, the above commentary still holds for the general case.

        # Pre-minified simulation duration in seconds.
        # sim_time_total_old = self.p.sim_time_total

        # event_cut = self.p.conf['cutting event']
        # event_cut['cut time'] = (
        #     self.p.sim_time_total * event_cut['cut time'] /
        #     sim_time_total_old)

        # Minify the physical dimensions of the cell cluster in meters. By
        # experimentation, the default simulation configuration both:
        #
        # * Exhibits expected instabilities for physical dimensions less than
        #   100e-6, typically due to the cell cluster cutting event. Since these
        #   instabilities are expected and hence best ignored, the dimensions
        #   specified below should *NEVER* be smaller than 100e-6.
        # * Exhibits unexpected instabilities for physical dimensions greater
        #   than 250e-6, typically due to non-linear interactions between
        #   channel electroosmosis and physical deformation. Since these
        #   instabilities are unexpected and hence *NOT* safely ignorable, the
        #   dimensions specified below should *ALWAYS* be in this range.
        #
        # Hence, these dimensions reside in an optimal range both avoiding
        # expected instabilities *AND* exposing unexpected instabilities.
        self.environment_size = min(self.environment_size, 100e-6)

        #FIXME: Uncommenting the following line reliably causes the following
        #test to fail with a computational instability during the "sim" phase:
        #
        #    ./test -k test_cli_sim_ecm
        #
        #Since it's unclear whether this behaviour is expected or not, this line
        #remains commented out. (If this behaviour is indeed unexpected, this
        #line should be shifted into the enable_exports_all() method and the
        #more preferable global default retained above).
        # self.environment_size = min(self.environment_size, 250e-6)

        # Minify ECM-specific grid size. For similar reasons as above, the
        # computational grid size specified below appears to be a hard minimum.
        self.p.grid_size = min(self.p.grid_size, 20)

        # Log this minification.
        logs.log_debug(
            'Minifying simulation to init '
            'for %f s at [ %f | %f ] s and sim '
            'for %f s at [ %f | %f ] s...',
            self.p.init_time_step,
            self.p.init_time_sampling,
            self.p.init_time_total,
            self.p.sim_time_step,
            self.p.sim_time_sampling,
            self.p.sim_time_total,
        )

    # ..................{ DISABLERS                          }..................
    def disable_interaction(self) -> None:
        '''
        Disable all simulation configuration options either requiring
        interactive user input _or_ displaying graphical output intended for
        interactive user consumption (e.g., plots, animations).

        This method is intended to be called by non-interactive automation
        (e.g., tests, scripts) expecting simulations to behave silently.
        '''

        self.p.anim.is_after_sim_show = False
        self.p.anim.is_while_sim_show = False
        self.p.plot.is_after_sim_show = False


    def disable_visuals(self) -> None:
        '''
        Disable all visual exports, including displaying and saving of all in-
        and post-simulation plots and animations.
        '''

        self.p.anim.is_after_sim = False
        self.p.anim.is_while_sim = False
        self.p.plot.is_after_sim = False

    # ..................{ ENABLERS                           }..................
    def enable_networks(self) -> None:
        '''
        Enable both biochemical reaction and gene regulatory networks.
        '''

        self.p.conf['gene regulatory network settings'][
            'gene regulatory network simulated'] = True

    # ..................{ ENABLERS ~ export                  }..................
    @type_check
    def enable_anim_video(self, writer_name: str, filetype: str) -> None:
        '''
        Enable encoding of all enabled animations as compressed video of the
        passed filetype with the preferred matplotlib animation writer of the
        passed name.

        Parameters
        ----------
        writer_name : str
            Name of the matplotlib animation writer with which to encode video
            (e.g., ``ffmpeg``, ``imagemagick``).
        filetype : str
            Filetype of videos to encode with this writer (e.g., ``mkv``).
        '''

        # Enable animations and animation saving in the general sense.
        self.enable_visuals_save()

        # Localize nested dictionaries for convenience.
        video = self.p.conf['results options']['save']['animations']['video']

        # Enable encoding of the passed filetype with the passed writer type.
        # For determinism, mandate that *ONLY* this writer (rather than two or
        # more writers) be used to do so.
        video['enabled'] = True
        video['filetype'] = filetype
        video['writers'] = [writer_name,]


    def enable_visuals_save(self) -> None:
        '''
        Enable saving of all visual exports, including all in- and
        post-simulation plots and animations.

        This method does *not* enable specific visual exports or simulation
        features required by specific visual exports.
        '''

        self.p.anim.is_while_sim_save = True
        self.p.anim.is_after_sim_save = True
        self.p.plot.is_after_sim_save = True

    # ..................{ ENABLERS ~ solver : fast           }..................
    def enable_solver_fast_exports(self) -> None:
        '''
        Enable all possible exports (e.g., CSVs, plots, animations) supported
        by the equivalent circuit-based BETSE solver excluding those requiring
        extracellular spaces, all features required by these exports, and any
        additional features trivially enabled *without* increasing time or
        space complexity.
        '''

        # Enable all simulation features, including the full BETSE solver but
        # excluding extracellular spaces.
        self._enable_solver_fast_features()

        # Disable extracellular spaces.
        self.p.is_ecm = False

        # Enable all possible exports excluding those requiring the full
        # solver.
        self._enable_exports(requirements_omit=phasereqs.SOLVER_FULL)


    def _enable_solver_fast_features(self) -> None:
        '''
        Enable all simulation features required by all exports (e.g., CSVs,
        plots, animations) supported by the equivalent circuit-based BETSE
        solver, excluding extracellular spaces.

        This method additionally enables optional settings improving test
        coverage but *not* explicitly required by these exports. Specifically,
        this method enables:

        * The full BETSE solver.
        * Saving of all visual exports.
        '''

        # Enable the equivalent circuit-based solver.
        self._enable_solver_fast()

        # Enable saving of these exports.
        self.enable_visuals_save()


    def _enable_solver_fast(self) -> None:
        '''
        Enable the equivalent circuit-based BETSE solver *and* disable all
        simulation features unsupported by this solver.
        '''

        self.p.solver_type = SolverType.FAST

    # ..................{ ENABLERS ~ solver : full           }..................
    def enable_solver_full_vg_ions(self) -> None:
        '''
        Enable all voltage-gated ion channels (e.g., sodium, potassium) *and*
        all features required by these channels.

        This method is intended to be called by non-interactive test suites
        exercising these channels. Specifically, this method enables:

        * The full BETSE solver.
        * The extracellular matrix (ECM).
        * The mammalian ion profile (i.e., ``animal``), enabling all ions.
        * The intervention increasing the permeability of all cell membranes to
          sodium (Na+).
        * The voltage-gated sodium (Na+) channel ``Nav1p2``, corresponding to
          the adult human brain.
        * The voltage-gated potassium (K+) channel ``K_Slow``.
        * Decreased time step and sampling rates, ensuring simulation stability.
        * Increased duration and cell count, exposing simulation instabilities.

        For efficiency, this method disables all visuals -- including both in-
        and post-simulation animations and plots.
        '''

        # For efficiency, disable all visuals.
        self.disable_visuals()

        # Enable all features required by these channels.
        self._enable_solver_full()
        self.p.is_ecm = True
        self.p.ion_profile = IonProfileType.MAMMAL

        # For stability, decrease both the time step and sampling rates.
        self.p.sim_time_step     = 1e-4
        self.p.sim_time_sampling = 1e-3

        # For completeness, increase both the duration and cell count.
        self.p.sim_time_total = 50e-3
        self.environment_size = 250e-6

        # Enable the intervention increasing sodium membrane permeability.
        # Although the current default values for this intervention track those
        # defined below fairly closely, the latter are nonetheless explicitly
        # defined below to avoid issues when the former inevitably change.
        sodium_membrane_permeability = self.p.conf['change Na mem']
        sodium_membrane_permeability['event happens'] = True
        sodium_membrane_permeability['change rate']   =  1.0e-3
        sodium_membrane_permeability['change start']  =  5.0e-3
        sodium_membrane_permeability['change finish'] = 30.0e-3
        sodium_membrane_permeability['apply to'] = ['Spot',]

        #FIXME: Refactor this to use the new networks formalism.
        # # Enable the voltage-gated sodium (Na+) channel Nav1p2.
        # voltage_gated_sodium_channel = self.p.conf['voltage gated Na+']
        # voltage_gated_sodium_channel['turn on'] = True
        # voltage_gated_sodium_channel['channel type'] = ['Nav1p2',]
        # # voltage_gated_sodium_channel['max value'] = 5.0e-6
        # voltage_gated_sodium_channel['apply to'] = ['base',]
        #
        # # Enable the voltage-gated potassium (K+) channel K_Slow.
        # voltage_gated_potassium_channel = self.p.conf['voltage gated K+']
        # voltage_gated_potassium_channel['turn on'] = True
        # voltage_gated_potassium_channel['channel type'] = ['K_Slow',]
        # # voltage_gated_potassium_channel['max value'] = 5.0e-7
        # voltage_gated_potassium_channel['apply to'] = ['base',]


    def _enable_solver_full(self) -> None:
        '''
        Enable the complete BETSE solver.
        '''

        self.p.solver_type = SolverType.FULL

    # ..................{ ENABLERS ~ solver : full : exports }..................
    def enable_solver_full_exports_ecm(self) -> None:
        '''
        Enable all possible exports (e.g., CSVs, plots, animations) supported
        by the full BETSE solver including those requiring extracellular spaces,
        all features required by these exports, and any additional features
        trivially enabled *without* increasing time or space complexity.
        '''

        # Enable all simulation features, including the full BETSE solver but
        # excluding extracellular spaces.
        self._enable_solver_full_features()

        # Enable extracellular spaces.
        self.p.is_ecm = True

        # Enable all possible exports.
        self._enable_exports()


    def enable_solver_full_exports_noecm(self) -> None:
        '''
        Enable all possible exports (e.g., CSVs, plots, animations) supported
        by the full BETSE solver excluding those requiring extracellular spaces,
        all features required by these exports, and any additional features
        trivially enabled *without* increasing time or space complexity.
        '''

        # Enable all simulation features, including the full BETSE solver but
        # excluding extracellular spaces.
        self._enable_solver_full_features()

        # Disable extracellular spaces.
        self.p.is_ecm = False

        # Enable all possible exports excluding those requiring extracellular
        # spaces.
        self._enable_exports(requirements_omit=phasereqs.ECM)

    # ..................{ PRIVATE ~ enablers                 }..................
    @type_check
    def _enable_exports(
        self,

        # Optional parameters.
        requirements_omit: SimPhaseRequirementsOrNoneTypes = None,
    ) -> None:
        '''
        Enable all possible exports (e.g., CSVs, plots, animations) excluding
        those requiring one or more of the passed simulation phase requirements
        (e.g., the full BETSE solver, extracellular spaces).

        Exports already enabled by the current simulation configuration are
        gracefully preserved as is rather than erroneously readded to their
        respective pipelines.

        Parameters
        ----------
        requirements_omit : SimPhaseRequirementsOrNoneTypes
            Immutable set of all simulation phase requirements such that
            exports requiring one or more requirements in this set are *not*
            enabled by this method. Defaults to ``None``, in which case all
            possible exports are unconditionally enabled.
        '''

        # Defer heavyweight imports.
        from betse.science.config.export.confexpabc import SimConfExportABC
        from betse.util.type.obj import objtest

        # Log this action.
        logs.log_debug('Analyzing pipeline exporters...')

        # Simulation phase encapsulating this configuration. Since this phase
        # *ONLY* serves as a thin wrapper around this configuration, the type
        # of phase is irrelevant; ergo, this type arbitrarily defaults to the
        # first phase type: seed.
        phase = SimPhase(kind=SimPhaseKind.SEED, p=self.p)

        # Default the set of requirements to omit to the empty set.
        if requirements_omit is None:
            requirements_omit = phasereqs.NONE

        # For each export pipeline...
        for pipe_export in SimPipesExport().PIPES_EXPORT:
            # Log this pipeline.
            logs.log_debug(
                'Analyzing pipeline "%s" exporters...', pipe_export.name)

            # Sequence of all export subconfigurations for this pipeline.
            pipe_exporters_conf = pipe_export.iter_runners_conf(phase)

            # Remove all export subconfigurations from this sequence, enabling
            # new test-specific subconfigurations to be added to this sequence
            # below without needless concern over conflicts and redundancy.
            pipe_exporters_conf.clear()

            # For each possible export supported by this pipeline (including
            # those currently disabled)...
            for pipe_exporter_metadata in pipe_export.iter_runners_metadata():
                # Log this iteration.
                logs.log_debug(
                    'Analyzing pipeline "%s" exporter "%s"...',
                    pipe_export.name, pipe_exporter_metadata.kind)

                # If this export requires one or more simulation features
                # omitted by the caller...
                if pipe_exporter_metadata.requirements.isintersection(
                    requirements_omit):
                    # Log this exclusion.
                    logs.log_debug(
                        'Excluding pipeline "%s" exporter "%s", '
                        'due to unsatisfied test requirements...',
                        pipe_export.name, pipe_exporter_metadata.kind)

                    # Continue to the next export.
                    continue
                # Else, this export requires no omitted simulation features.

                # Log this inclusion.
                logs.log_debug(
                    'Including pipeline "%s" exporter "%s"...',
                    pipe_export.name, pipe_exporter_metadata.kind)

                # Create and append a new default subconfiguration of this
                # export to the sequence of these subconfigurations.
                pipe_exporter_conf = pipe_exporters_conf.append_default()

                # If this is *NOT* actually an export subconfiguration, raise
                # an exception.
                objtest.die_unless_instance(
                    obj=pipe_exporter_conf, cls=SimConfExportABC)

                # Copy across the type of this export subconfiguration.
                pipe_exporter_conf.kind = pipe_exporter_metadata.kind

    # ..................{ PRIVATE ~ enablers : solver        }..................
    def _enable_solver_full_features(self) -> None:
        '''
        Enable all simulation features required by all exports (e.g., CSVs,
        plots, animations) -- including the full BETSE solver but excluding
        extracellular spaces.

        This method additionally enables optional settings improving test
        coverage but *not* explicitly required by these exports. Specifically,
        this method enables:

        * The full BETSE solver.
        * Saving of all visual exports.
        * Cell enumeration, labelling each cell by its 0-based index.
        * Current overlays, displaying current density streamlines.
        * All ion concentration plots and animations (e.g., calcium (Ca+),
          hydrogen (H+; pH)) by enabling:

          * The mammalian ion profile (i.e., ``animal``), enabling all ions.

        * The deformation plot and animation by enabling:

          * Galvanotaxis (i.e., deformations).

        * The ``fluid_intra`` plot and animation by enabling:

          * Fluid flow.

        * The ``pressure_mechanical`` plot and animation by enabling:

          * The mechanical pressure event.
        * The ``pressure_osmotic`` plot and animation by enabling:

          * Osmotic pressure.

        * The ``pump_density`` plot and animation by enabling:

          * Channel electroosmosis.

        * The ``voltage_polarity`` plot and animation by enabling:

          * Cell polarizability.

        * All other plots and animations *not* requiring extracellular spaces.
        '''

        # Enable the full solver.
        self._enable_solver_full()

        # Enable saving of these exports.
        self.enable_visuals_save()

        # Localize nested dictionaries for convenience.
        results = self.p.conf['results options']
        variable = self.p.conf['variable settings']

        # Enable all simulation features required by these exports.
        self.p.ion_profile = IonProfileType.MAMMAL
        self.p.conf['apply pressure']['event happens'] = True
        # variable['channel electroosmosis']['turn on'] = True  # This feature has been removed
        variable['deformation']['turn on'] = True
        # variable['fluid flow']['include fluid flow'] = True
        variable['pressures']['include osmotic pressure'] = True
        self.p.cell_polarizability = 1e-4

        # Enable all optional settings supported by these exports.
        results['visuals']['cell indices']['show'] = True
        results['overlay currents'] = True



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Layer subclasses spatially shading the cell cluster as a continuous surface.
'''

# ....................{ IMPORTS                            }....................
from betse.science.math import mathunit
from betse.science.visual.layer.vector.lyrvecabc import (
    LayerCellsVectorColorfulABC)
from betse.util.type.types import type_check, IterableTypes, SequenceOrNoneTypes

# ....................{ SUBCLASSES                         }....................
class LayerCellsVectorSmoothGrids(LayerCellsVectorColorfulABC):
    '''
    Layer subclass plotting a single vector spatially situated at environmental
    grid space centres (e.g., extracellular electric field magnitudes) as a
    continuous surface depicted by a polygonal mesh onto the cell cluster for
    one on more simulation steps.

    Attributes
    ----------
    _surface_image : matplotlib.image.AxesImage
        Surface image plot of this vector for the current time step.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Default all instance variables.
        self._surface_image = None

    # ..................{ SUPERCLASS                         }..................
    @property
    def color_data(self) -> SequenceOrNoneTypes:
        return self._vector.times_grids_centre


    @type_check
    def _layer_first_color_mappables(self) -> IterableTypes:

        # Surface image plot of these magnitudes plotted for this time step. See
        # the matplotlib.axes.imshow() docstring for further details.
        self._surface_image = self._visual.axes.imshow(
            # Two-dimensional array of all grid data for this time step,
            # spatially situated at environmental grid space centres.
            X=self._vector.times_grids_centre[self._visual.time_step],
            # self._current_density_magnitude_time_series[self._visual.time_step],

            # Colormap converting input data values into output color values.
            cmap=self._visual.colormap,

            # X and Y coordinates of the boundaries of this environmental grid.
            extent=self._phase.cache.upscaled.extent,

            # The [0, 0] index of this two- or three-dimensional Numpy array
            # resides at the lower-left corner of this figure's axes.
            origin='lower',

            # Z-order of this plot with respect to other artists.
            zorder=self._zorder,
        )
        # self._surface_image.set_clim(0, 152)

        # Map this surface image plot onto the figure colorbar, returned as a
        # 1-tuple to comply with the superclass API.
        return (self._surface_image,)


    def _layer_next(self) -> None:

        # print('Here!')
        # self._visual._rescale_color_mappables()

        # Replace all obsoleted grid data plotted for the prior time step by the
        # updated grid data for this time step.
        self._surface_image.set_data(
            # self._current_density_magnitude_time_series[-1])
            # self._current_density_magnitude_time_series[self._visual.time_step])
            self._vector.times_grids_centre[self._visual.time_step])


class LayerCellsVectorSmoothRegions(LayerCellsVectorColorfulABC):
    '''
    Layer subclass plotting a single vector spatially situated at Voronoi region
    centres (e.g., transmembrane voltage averages) as a continuous
    Gouraud-shaded surface depicted by a polygonal mesh onto the cell cluster
    for one on more simulation steps, interpolating the cell data for each cell
    across the smooth spatial continuum inhabited by that cell.

    Attributes
    ----------
    _cluster_tri_mesh : matplotlib.collections.TriMesh
        Unstructured triangulation mesh grid interpolating the cell data for
        each cell over the cell cluster as a contiuous whole.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Default all instance attributes.
        self._cluster_tri_mesh = None

    # ..................{ SUPERCLASS                         }..................
    @property
    def color_data(self) -> SequenceOrNoneTypes:

        return self._vector.times_regions_centre


    def _layer_first_color_mappables(self) -> IterableTypes:

        # X and Y coordinates of all cell membrane midpoints.
        membranes_midpoint_x = mathunit.upscale_coordinates(
            self._phase.cells.mem_mids_flat[:,0])
        membranes_midpoint_y = mathunit.upscale_coordinates(
            self._phase.cells.mem_mids_flat[:,1])

        # Membrane midpoint-centred data for this time step.
        membranes_midpoint_data = self._vector.times_membranes_midpoint[
            self._visual.time_step]

        # Gouraud-shaded triangulation mesh for this cell cluster, computed from
        # the Delaunay hull of the non-triangular centers of these regions.
        self._cluster_tri_mesh = self._visual.axes.tripcolor(
            # Positional arguments. Thanks to internal flaws in the
            # matplotlib.tri.tripcolor() function parsing arguments passed
            # to the matplotlib.axes.tripcolor() method called here, the first
            # four arguments *MUST* be passed as positional arguments.
            membranes_midpoint_x, membranes_midpoint_y, membranes_midpoint_data,

            # Keyword arguments. All remaining arguments *MUST* be passed as
            # keyword arguments.
            shading='gouraud',
            vmin=self._visual.color_min,
            vmax=self._visual.color_max,

            # Colormap converting input data values into output color values.
            cmap=self._visual.colormap,

            # Z-order of this mesh with respect to other artists.
            zorder=self._zorder,
        )

        # Map this triangulation mesh onto the figure colorbar, returned as a
        # 1-tuple to comply with the superclass API.
        return (self._cluster_tri_mesh,)


    def _layer_next(self) -> None:

        # Membrane midpoint-centred data for this time step.
        membranes_midpoint_data = self._vector.times_membranes_midpoint[
            self._visual.time_step]

        # Gouraud-shade this triangulation mesh with these color values.
        self._cluster_tri_mesh.set_array(membranes_midpoint_data)


#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Sentinel** (i.e., objects of arbitrary unique value, typically used to
distinguish edge-case results from the standard ``None`` singleton) facilities.
'''

# ....................{ CLASSES                           }....................
class Sentinel(object):
    '''
    Class encapsulating sentinel objects of arbitrary (albeit unique) value.

    Instances of this class are intended to be used as placeholder objects in
    iterables, typically to identify erroneous or edge-case algorithm input.
    '''

    def __repr__(self) -> str:
        '''
        Human- and machine-readable representation of this sentinel.

        This method has been overridden purely to improve the debuggability of
        algorithms requiring instances of this class.
        '''

        return 'Sentinel()'

# ....................{ CONSTANTS                         }....................
SENTINEL = Sentinel()
'''
Sentinel object of arbitrary value.

This object is internally leveraged by various utility functions (e.g.,
:func:`betse.util.type.iterable.iterables.zip_isometric`) to identify erroneous
and edge-case input (e.g., iterables of insufficient length).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Matplotlib-specific **z-order** (i.e., positive integers ordering artist
drawing, such that artists with larger z-orders are drawn over artists with
smaller z-orders) facilities.

See Also
----------
http://matplotlib.org/examples/pylab_examples/zorder_demo.html
    Canonical z-order example from which most constants defined by this
    submodule were derived.
'''

# ....................{ CONSTANTS                          }....................
ZORDER_PATCH = 1
'''
Default **z-order** (i.e., positive integer ordering artist drawing, such that
artists with larger z-orders are drawn over artists with smaller z-orders) for
patch artists (e.g., :class:`Patch`, :class:`PatchCollection`).

This is the lowest default z-order, thus drawing patch artists under all other
artists by default.
'''


ZORDER_LINE = 2
'''
Default **z-order** (i.e., positive integer ordering artist drawing, such that
artists with larger z-orders are drawn over artists with smaller z-orders) for
line artists (e.g., `Line2D`, `LineCollection`, `StreamplotSet`).

This is the middle default z-order, thus drawing line artists over all patch
artists but under all text artists by default.
'''


ZORDER_TEXT = 3
'''
Default **z-order** (i.e., positive integer ordering artist drawing, such that
artists with larger z-orders are drawn over artists with smaller z-orders) for
text artists (e.g., `Text`).

This is the highest default z-order, thus drawing text artists over all other
artists by default.
'''


ZORDER_STREAM = (ZORDER_LINE + ZORDER_TEXT) / 2
'''
BETSE-specific **z-order** (i.e., positive integer ordering artist drawing,
such that artists with larger z-orders are drawn over artists with smaller
z-orders) for streamplots (e.g., `StreamplotSet`).

This magic number has been chosen such that streamplots with this z-order will
be drawn over all line and patch artists but under all text artists by default.
Streamplots are technically a variant of line artists but sufficiently non-
linear (and visually busy) to warrant separate handling.

It may also be pertinent to note that recent Matplotlib releases as of this
writing (1.40) accidentally broke backward compatibility with respect to
default streamplot z-order. Specifically, streamplots were assigned a default
z-order of `ZORDER_PATCH` rather than `ZORDER_LINE`:

    https://github.com/matplotlib/matplotlib/pull/5567
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level versioning facilities.

Design
------
For generality, most callables defined by this submodule accept `PEP
440-compliant`_ version specifiers in any of several permissible formats. As
encapsulated by the :data:`VersionTypes` tuple, these are:

* :class:`str`, specifying versions in ``.``-delimited positive integer format
  (e.g., ``2.4.14.2.1.356.23``).
* :class:`tuple`, specifying versions as one or more positive integers (e.g.,
  ``(2, 4, 14, 2, 1, 356, 23)``),
* :class:`VersionSetuptoolsTypes`, specifying versions as instance variables
  convertible into both of the prior formats (e.g.,
  ``VersionSetuptoolsTypes('2.4.14.2.1.356.23')``).

.. _PEP 440-compliant:
   https://www.python.org/dev/peps/pep-0440
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import Tuple
from betse.exceptions import BetseVersionException
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.types import RegexCompiledType, VersionTypes
from collections.abc import Collection as CollectionABC

# ....................{ TESTERS ~ greater                  }....................
def is_greater_than(version_1: VersionTypes, version_2: VersionTypes) -> bool:
    '''
    ``True`` only if the first passed version is strictly greater than the
    second passed version.

    Parameters
    ----------
    version_1 : VersionTypes
        First version (e.g., ``2.0.1``, ``(2, 0, 1)``) to be tested.
    version_2 : VersionTypes
        Second version (e.g., ``1.0.2``, ``(1, 0, 2)``) to be tested.

    Returns
    -------
    bool
        ``True`` only if the first version is greater than the second version.

    Raises
    ------
    BetseVersionException
        If either version is *not* `PEP 440-compliant`_.

    .. _PEP 440-compliant:
       https://www.python.org/dev/peps/pep-0440
    '''

    # Nothing is hard. Everything is easy.
    return to_comparable(version_1) > to_comparable(version_2)


def is_greater_than_or_equal_to(
    version_1: VersionTypes, version_2: VersionTypes) -> bool:
    '''
    ``True`` only if the first passed version is greater than or equal to the
    second passed version.

    Parameters
    ----------
    version_1 : VersionTypes
        First version (e.g., ``2.0.1``, ``(2, 0, 1)``) to be tested.
    version_2 : VersionTypes
        Second version (e.g., ``1.0.2``, ``(1, 0, 2)``) to be tested.

    Returns
    -------
    bool
        ``True`` only if the first version is greater than or equal to the
        second version.

    Raises
    ------
    BetseVersionException
        If either version is *not* `PEP 440-compliant`_.

    .. _PEP 440-compliant:
       https://www.python.org/dev/peps/pep-0440
    '''

    # Nothing is easy, until it is.
    return to_comparable(version_1) >= to_comparable(version_2)

# ....................{ TESTERS ~ less                     }....................
def is_less_than(version_1: VersionTypes, version_2: VersionTypes) -> bool:
    '''
    ``True`` only if the first passed version is strictly less than the second
    passed version.

    Parameters
    ----------
    version_1 : VersionTypes
        First version (e.g., ``2.0.1``, ``(2, 0, 1)``) to be tested.
    version_2 : VersionTypes
        Second version (e.g., ``1.0.2``, ``(1, 0, 2)``) to be tested.

    Returns
    -------
    bool
        ``True`` only if the second version is less than the first version.

    Raises
    ------
    BetseVersionException
        If either version is *not* `PEP 440-compliant`_.

    .. _PEP 440-compliant:
       https://www.python.org/dev/peps/pep-0440
    '''

    # Some things are hard. This is not those things.
    return to_comparable(version_1) < to_comparable(version_2)


def is_less_than_or_equal_to(
    version_1: VersionTypes, version_2: VersionTypes) -> bool:
    '''
    ``True`` only if the first passed version is less than or equal to the
    second passed version.

    Parameters
    ----------
    version_1 : VersionTypes
        First version (e.g., ``2.0.1``, ``(2, 0, 1)``) to be tested.
    version_2 : VersionTypes
        Second version (e.g., ``1.0.2``, ``(1, 0, 2)``) to be tested.

    Returns
    -------
    bool
        ``True`` only if the second version is less than or equal to the first
        version.

    Raises
    ------
    BetseVersionException
        If either version is *not* `PEP 440-compliant`_.

    .. _PEP 440-compliant:
       https://www.python.org/dev/peps/pep-0440
    '''

    # Everything is easy, until it isn't.
    return to_comparable(version_1) <= to_comparable(version_2)

# ....................{ CONVERTERS                         }....................
def to_comparable(version: VersionTypes) -> Tuple[int, ...]:
    '''
    Passed version converted into a **comparable version type** (i.e., type
    suitable for use both as parameters to callables accepting arbitrary
    version specifiers *and* as operands to numeric operators comparing such
    specifiers).

    Specifically, if the passed version is of type:

    * :class:`str`:

      * This string is munged to comply with industry-standard semantics.
        Specifically:

        * All tildes (i.e., ``~`` characters) in this string are globally
          replaced with hyphens (i.e., ``-`` characters). For unknown reasons,
          older (and possibly current) implementations of the low-level
          :func:`pkg_resources.parse_version` function underlying this
          higher-level function sort versions containing tilde characters as
          strictly less than versions *not* containing tilde characters (e.g.,
          sorting ``5.9.0~a1`` as less than ``5.7.0``). This is blatantly
          wrong, of course. Since version strings associated with dependencies
          often contain tilde characters, failure to address this subtle issue
          would induce subtle issues elsewhere throughout this application.

      * A new :class:`VersionSetuptoolsTypes` instance is instantiated and
        returned from this version.

    * :class:`tuple`, a new :class:`VersionSetuptoolsTypes` instance is
      instantiated and returned from this version.
    * :class:`VersionSetuptoolsTypes`, the same
      :class:`VersionSetuptoolsTypes` is returned as is.

    Caveats
    -------
    The version returned by this function is *only* safely comparable with
    versions of the same type. In particular, the
    :class:`VersionSetuptoolsTypes` type does *not* necessarily support direct
    comparison with either the :class:`tuple` *or* `class:`str` version types;
    :class:`VersionSetuptoolsTypes` supported both under older but *not* newer
    versions of :mod:`setuptools`. *shakes fist*

    Parameters
    ----------
    version : VersionTypes
        Version (e.g., ``2.0.1``, ``(2, 0, 1)``) to be converted.

    Returns
    -------
    VersionSetuptoolsTypes
        :mod:`setuptools`-specific version converted from this version.

    Raises
    ------
    BetseVersionException
        If this version is *not* `PEP 440-compliant`_.

    .. _PEP 440-compliant:
       https://www.python.org/dev/peps/pep-0440

    Examples
    --------
    .. code-block:: python

       >>> from betse.util.type.numeric import versions
       >>> versions.to_comparable('2.4.14.2.1.356.23')
       Version('2.4.14.2.1.356.23')
       >>> versions.to_comparable((2, 4, 14, 2, 1, 356, 23))
       (2, 4, 14, 2, 1, 356, 23)
    '''

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from betse.util.type.text.regexes import replace_substrs

    # ....................{ LOCALS                         }....................
    # Version passed to this function, preserved for subsequent embedding in
    # human-readable exception messages.
    version_old = version

    # ....................{ STRING                         }....................
    # Note that there are *MANY* approaches for converting strings into
    # comparable versions, including:
    # * The packaging.version.parse() function, whose API resembles that of
    #   setuptools called below. In theory, the "packaging" package *SHOULD* be
    #   a setuptools dependency and hence always importable wherever the
    #   "pkg_resources" and "setuptools" packages are importable. In practice,
    #   this does *NOT* appear to be the case.
    # * The "distutils.version.LooseVersion" class, whose API resembles that of
    #   setuptools called below. However, whereas the latter API is PEP
    #   440-conformant, the former API is not.
    # * The setuptools.pkg_resources.parse_version() function previously called
    #   below. As "setuptools" has since deprecated "pkg_resources", however,
    #   deferring to that function is no longer a reasonable proposition.

    # If this is a string-formatted version...
    if isinstance(version, str):
        # Compiled regular expression matching all invalid delimiters (e.g.,
        # "a", "b", "rc") in "."-delimited version specifiers.
        VERSION_INVALID_DELIMITER_REGEX = _get_version_invalid_delimiter_regex()

        # Sanitized version, replacing each invalid delimiter in this
        # unsanitized version with the valid delimiter ".".
        version_sanitized = replace_substrs(
            text=version,
            regex=VERSION_INVALID_DELIMITER_REGEX,
            replacement='.',
        )

        # List of one or more string-formatted version components split from
        # this sanitized version on "." delimiters. Note that this behaves as
        # expected when this version contains *NO* "." delimiters.
        version = version_sanitized.split('.')
    assert isinstance(version, CollectionABC), (
        f'{repr(version)} not version collection.')

    # ....................{ TUPLE                          }....................
    # Attempt to coerce this collection of possibly non-integer objects (e.g.,
    # string-formatted version components) into a tuple of one or more
    # integer-formatted version components.
    try:
        version = tuple(int(version_item) for version_item in version)
    except Exception as exception:
        raise BetseVersionException(
            f'PEP 440 version {repr(version_old)} invalid '
            f'(i.e., contains one or more characters that are neither '
            f'integers nor "." delimiters).'
        ) from exception

    # Return this tuple-formatted version as is.
    return version

# ....................{ PRIVATE ~ getters                  }....................
@func_cached
def _get_version_invalid_delimiter_regex() -> RegexCompiledType:
    '''
    Compiled regular expression matching all **invalid delimiters** in
    ``"."``-delimited version specifiers.

    Invalid delimiters typically denote pre-release "sub-versions," including
    such substrings as:

    * The ``"a"`` in ``"149.0a5"``.
    * The ``"b"`` in ``"1.3.0b0"``.
    * The ``"rc"`` in ``"0.19.0rc1"``.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.regexes import compile_regex

    # Create, return, and cache this expression.
    return compile_regex(
        # One or more instances of an invalid delimiter, defined as...
        r'(?:'
        # Dismantled, this is:
        # * "a", a single character signifying an alpha pre-release.
        # * "b", a single character signifying a beta pre-release.
        # * "~", a PySide2-specific delimiter intended to signify a "-". Look.
        #   Don't ask. Just accept that the real world is a nasty place.
        # * "-", a non-standard delimiter in some version specifiers.
        r'[ab~-]|'
        # Release candidate pre-release suffix.
        r'rc'
        r')+'
    )



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all matplotlib-based layer subclasses spatially
plotting vector fields onto the cell cluster.
'''

# ....................{ IMPORTS                            }....................
from betse.science.math.vector.vecfldcls import VectorFieldCellsCache
from betse.science.visual.layer.lyrabc import (
    LayerCellsABC, LayerCellsColorfulABC)
from betse.util.type.types import type_check

# ....................{ SUBCLASSES                         }....................
class LayerCellsFieldColorlessABC(LayerCellsABC):
    '''
    Abstract base class of all classes spatially plotting vector fields of
    arbitrary data onto the cell cluster (e.g., current density, electric field)
    such that neither the components nor magnitudes of these fields are mappable
    as colors onto the colorbars of parent plots and animations.

    Attributes
    ----------
    _field : VectorFieldCellsCache
        Cache of various vector fields of the same underlying data situated
        along different coordinate systems for all time steps to be animated.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, field: VectorFieldCellsCache) -> None:
        '''
        Initialize this layer.

        Parameters
        ----------
        field : VectorFieldCellsCache
            Cache of various vector fields of the same underlying data situated
            along different coordinate systems for all time steps to be
            animated.
        '''

        # Initialize our superclass.
        super().__init__()

        # Classify all passed parameters.
        self._field = field


class LayerCellsFieldColorfulABC(LayerCellsColorfulABC):
    '''
    Abstract base class of all classes spatially plotting vector fields of
    arbitrary data onto the cell cluster (e.g., current density, electric field)
    such that the components and/or magnitudes of these fields are mappable as
    colors onto the colorbars of parent plots and animations.

    Attributes
    ----------
    _field : VectorFieldCellsCache
        Cache of various vector fields of the same underlying data situated
        along different coordinate systems for all time steps to be animated.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, field: VectorFieldCellsCache) -> None:
        '''
        Initialize this layer.

        Parameters
        ----------
        field : VectorFieldCellsCache
            Cache of various vector fields of the same underlying data situated
            along different coordinate systems for all time steps to be
            animated.
        '''

        # Initialize our superclass.
        super().__init__()

        # Classify all passed parameters.
        self._field = field



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **decorator-based profiling** (i.e., debugging reports of the space
and/or time complexity of decorated callables) facilities.
'''

# ....................{ IMPORTS                            }....................
import time
from betse.util.io.log import logs
from betse.util.type.types import (
    type_check, CallableTypes, CallableOrNoneTypes, StrOrNoneTypes)
from functools import wraps

# ....................{ DECORATORS                         }....................
# Decorators accepting optional arguments require additional... finesse.
@type_check
def log_time_seconds(
    # Callable to be decorated passed positionally if at all by Python itself.
    # If:
    #
    # * Passed, this function is being called as a decorator.
    # * Unpassed, this function is being called as a decorator factory expected
    #   to produce the decorator closure specific to the set of all
    #   subsequently passed optional parameters.
    func: CallableOrNoneTypes = None,
    # Require *ALL* subsequent optional parameters to be passed as
    # non-ambiguous keyword arguments rather than ambiguous positional
    # arguments. If *NOT* done, there would exist no means of deciding whether
    # this function is being called as a decorator or decorator factory in the
    # edge case of a decorator passed a callable as a first optional parameter.
    *,
    # Optional parameters passed to this function when called as a decorator
    # factory rather than as a decorator.
    noun: StrOrNoneTypes = None,
    verb: str = 'completed'
) -> CallableTypes:
    '''
    Decorate the passed callable (e.g., function, lambda, method) to log the
    **wall clock time** (i.e., cumulative time spent in both kernel- and
    userspace while in the body of this callable) of each call to this
    callable, denominated in possibly fractional seconds.

    Parameters
    ----------
    func : CallableOrNoneTypes
        Callable to be decorated if this function is being called as a
        decorator *or* ``None`` if this function is being called as a
        decorator factory.
    noun : StrOrNoneTypes
        Human-readable string describing this callable to be logged with the
        time spent in each call of this callable. If the first character of
        this string is *not* already uppercased, this function implicitly
        uppercases this character for readability. Defaults to ``None``, in
        which case a human-readable string describing this callable is
        defaulted to (e.g., ``Method skromt_og_kolabrenning()``).
    verb : str
        Human-readable string describing the action performed by this callable
        to be logged with the time spent in each call of this callable.
        Defaults to a general-purpose verb.

    See Also
    ----------
    https://stackoverflow.com/a/24617244/2809027
        StackOverflow answer inspiring this clever (albeit obtuse, admittedly)
        decorator design.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.call import callables
    from betse.util.type.text.string import strs

    # Decorator factory creating a decorator specific to the passed parameters.
    def _log_time_seconds_decorator(func: CallableTypes) -> CallableTypes:
        '''
        Create and return a closure decorating the passed callable to log the
        **wall clock time** (i.e., cumulative time spent in both kernel- and
        userspace while in the body of this callable) of each call to this
        callable, denominated in possibly fractional seconds.
        '''

        # Permit these enclosed variables from the outer scope to be modified.
        nonlocal noun

        # If unpassed, default this noun to the name of this callable.
        if noun is None:
            noun = callables.to_str(func)

        # Uppercase the first character of this string for readability.
        noun = strs.uppercase_char_first(noun)

        # Closure decorating this callable to be returned.
        @wraps(func)
        def _log_time_seconds_decorated(*args, **kwargs) -> object:

            # Current time in fractional seconds.
            start_time = time.time()

            # Call this function, passed all passed parameters and preserving the
            # return value as is.
            return_value = func(*args, **kwargs)

            # Cumulative time in fractional seconds spent in this call.
            end_time = time.time() - start_time

            # Log this time, rounded to two decimal places for readability.
            logs.log_info('%s %s in %.2f seconds.', noun, verb, end_time)

            # Return this return value.
            return return_value

        # Return this decorated callable.
        return _log_time_seconds_decorated

    # Return...
    return (
        # If this function is called as a decorator factory, this factory.
        _log_time_seconds_decorator if func is None else
        # Else, this function is called as a decorator. In this case, return a
        # decorator specific to the passed parameters.
        _log_time_seconds_decorator(func)
    )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **generator** (i.e., objects satisfying the standard generator API)
facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.util.type.types import type_check, GeneratorType, IterableTypes

# ....................{ GENERATORS                        }....................
def empty_generator() -> GeneratorType:
    '''
    Empty generator yielding... absolutely nothing.

    See Also
    ----------
    https://stackoverflow.com/a/36658865/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    yield from ()

# ....................{ TESTERS                           }....................
def is_generator(obj: object) -> bool:
    '''
    ``True`` only if the passed object is a generator.
    '''

    # Best things in life are free.
    return isinstance(obj, GeneratorType)

# ....................{ GETTERS                           }....................
@type_check
def get_length(generator: GeneratorType) -> int:
    '''
    Length of the passed **finite generator** (i.e., generator guaranteed to
    yield only a finite number of values).

    Caveats
    ----------
    **This function consumes the passed generator,** as a harmful (albeit
    unavoidable) side effect of calculating the length of this generator.

    **This function fails to halt if the passed generator is infinite,** as a
    harmful (albeit unavoidable) side effect of needing to entirely consume
    this generator in order to calculate its length.

    See Also
    ----------
    https://stackoverflow.com/a/31350424
        StackOverflow answer strongly inspiring this implementation.
    '''

    # Convert this generator into a list and return the length of this list.
    #
    # While clearly non-ideal, this approach is well-known to be the most
    # time-efficient solution. The most space-efficient solution is quite
    # different, but also approximately four times slower. Since time is scarce
    # and space is cheap in the general case, we prefer the sane approach.
    return len(list(generator))

# ....................{ CONVERTERS                        }....................
@type_check
def to_tuple_if_generator(iterable: IterableTypes) -> IterableTypes:
    '''
    Tuple of all items iteratively yielded by the passed iterable if this
    iterable is a generator *or* this iterable as is otherwise (i.e., if this
    iterable is *not* a generator).
    '''

    return tuple(iterable) if is_generator(iterable) else iterable



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level exception handling facilities.
'''

# ....................{ IMPORTS                           }....................
import traceback
from betse.util.type import types
from betse.util.type.types import type_check
from io import StringIO

# Private string constants required by our ad-hoc forwardport of the removed
# traceback._iter_chain() function. Fortuitously, these constants remain
# unchanged in Python >= 3.5.
from traceback import _cause_message, _context_message

# ....................{ GETTERS                           }....................
@type_check
def get_traceback(exception: Exception) -> str:
    '''
    Non-human-readable traceback associated with the passed exception,
    typically intended to be logged and/or displayed to end users for debugging
    purposes.

    Parameters
    ----------
    exception : Exception
        Exception to return this traceback for.

    Returns
    ----------
    str
        Non-human-readable traceback associated with this exception.
    '''

    # Terse synopsis and verbose traceback for this exception.
    _, exc_traceback = get_metadata(exception)

    # Return this traceback.
    return exc_traceback


@type_check
def get_metadata(exception: Exception) -> tuple:
    '''
    Tuple of various metadata specific to the passed exception, typically
    intended to be logged and/or displayed to end users.

    Parameters
    ----------
    exception : Exception
        Exception to return metadata for.

    Returns
    ----------
    (str, str)
        2-tuple ``(synopsis, traceback)``, where:
        * ``synopsis`` is this exception's human-readable synopsis.
        * ``traceback`` is this exception's non-human-readable traceback.
    '''

    # Avoid circular import dependencies.
    from betse.util.io.error import errhaiku
    from betse.util.py import pyident
    from betse.util.type.text import regexes
    from betse.util.type.text.string import strjoin, strs

    # Generator yielding 2-tuples "(exception, traceback)" for all parent
    # exceptions of this exception *AND* this exception (in that order), where
    # "exception" is each exception and "traceback" is the traceback stored for
    # each exception.
    #
    # Sadly, this list is only gettable via the private traceback._iter_chain()
    # function in older versions of Python. Since this function is unavailable
    # in newer versions of Python, an application-specific compatibility
    # function is called instead.
    exc_parents_generator = _iter_chain(exception, exception.__traceback__)

    # Tuple of 2-tuples "(exception, traceback)" in the reverse order yielded
    # by this generator, preserving readability by ensuring that this exception
    # is logged first, the parent exception of this exception (if any) is
    # logged second, and so forth.
    exc_parents = tuple(reversed(tuple(exc_parents_generator)))

    # 0-based index of the last exception in this list.
    exc_parent_last_index = len(exc_parents) - 1

    # String buffer containing a human-readable synopsis of each exception in
    # this chain, unconditionally output to stderr.
    exc_iota_buffer = StringIO()

    # String buffer containing a non-human-readable traceback of each exception
    # in this chain, conditionally logged to the logfile.
    exc_full_buffer = StringIO()

    # Human-readable header prefixing each such buffer.
    buffer_header = 'Exiting prematurely due to fatal error:\n\n'

    # Initialize these buffers to this header.
    exc_iota_buffer.write(buffer_header)
    exc_full_buffer.write(buffer_header)

    # For each parent exception and that exception's traceback...
    for exc_parent_index, (exc_parent, exc_parent_traceback) in (
        enumerate(exc_parents)):
        # If this exception is a string, append this string to the synopsis
        # buffer as is and continue to the next parent. This is an edge case
        # that should *NEVER* happen... but could.
        if types.is_str(exc_parent):
            exc_iota_buffer.write(exc_parent + '\n')
            continue

        # List of traceback lines, excluding this exception's message.
        exc_traceback_lines = traceback.format_exception(
            type(exc_parent),
            exc_parent,
            exc_parent_traceback)
        exc_traceback_lines.pop()

        # List of exception message lines, excluding traceback and hence
        # consisting only of this exception type and original message.
        exc_message_lines = traceback.format_exception_only(
            type(exc_parent), exc_parent)
        assert types.is_sequence_nonstr_nonempty(exc_message_lines), (
            types.assert_not_sequence_nonstr_nonempty(
                exc_message_lines, 'Exception message lines'))

        # If the exception type prefixing the last line of this message is
        # itself prefixed by the expected and hence ignorable fully-qualified
        # name of the subpackage defining BETSE exceptions, truncate this
        # prefix for brevity.
        #
        # Note that the format_exception_only() function guarantees the last
        # line of this message to *ALWAYS* be "the message indicating which
        # exception occurred."
        if exc_message_lines[-1].startswith('betse.exceptions.'):
            exc_message_lines[-1] = exc_message_lines[-1][
                len('betse.exceptions.'):]

        # Last line of this message. By design, the format_exception_only()
        # function guarantees this line to *ALWAYS* be "the message
        # indicating which exception occurred."
        exc_message_line = exc_message_lines[-1]

        # Append this message to the traceback buffer *BEFORE* appending a
        # truncation of this message to the message buffer.
        exc_full_buffer.write(strjoin.join(exc_message_lines))
        #print('exception string: '+ exc_message_lines[-1])

        # Split the last line of this message into a non-human-readable
        # exception class and ideally human-readable exception message. If this
        # exception is not "None" *AND* is convertable without raising
        # exceptions into a string, both format_exception_only() and
        # _format_final_exc_line() guarantee this line to be formatted as:
        #     "${exc_class}: ${exc_message}"
        exc_message_match_groups = regexes.get_match_groups_numbered(
            exc_message_line, r'^({})(?:\s*|:\s+(.+))$'.format(
                pyident.IDENTIFIER_QUALIFIED_REGEX))

        # This message is guaranteed to be prefixed by a class name.
        exc_class_name = exc_message_match_groups[0]

        # This message is *NOT* guaranteed to be prefixed by a non-empty
        # message (e.g., assert statements passed no message).
        #
        # If a non-empty message matched, prefer that.
        exc_message = None
        if exc_message_match_groups[1] is not None:
            exc_message = exc_message_match_groups[1]
        # Else if a debug assertion failed with no explicit message, prefer the
        # exception context directly detailing this assertion.
        elif exc_class_name == 'AssertionError':
            exc_message = 'Debug assertion failed: {}'.format(
                # A traceback line typically contains an internal newline. The
                # substring preceding this newline details the file and
                # function containing the corresponding call; the substring
                # following this newline is this call. Ignore the former.
                regexes.remove_substrs(
                    exc_traceback_lines[-1], r'^.+\n\s*'))
        # Else, convert this exception's class name into a human-readable
        # message (e.g., from "FileNotFoundError" to "File not found error.").
        # Well, try... at least!
        else:
            exc_message = strs.uppercase_char_first(
                pyident.convert_camelcase_to_whitespaced_lowercase(
                    exc_class_name))
        assert types.is_str_nonempty(exc_message), (
            types.assert_not_str_nonempty(
                exc_message, 'Exception message'))

        # If this class is "KeyError", this message is the single-quoted name
        # of a non-existent key in a dictionary whose access raised this
        # exception.  Replace this by a human-readable message.
        if exc_class_name == 'KeyError':
            exc_message = 'Dictionary key {} not found.'.format(
                exc_message)

        # Append this message to the synopsis buffer. For readability, this
        # message is wrapped to the default terminal width and each wrapped
        # line prefixed by indentation.
        exc_iota_buffer.write(strs.wrap(
            text=exc_message, line_prefix='    '))

        # If this exception has a traceback, append this traceback to the
        # traceback but *NOT* synopsis buffer.
        if exc_parent_traceback:
            # Append a traceback header.
            exc_full_buffer.write(
                '\nTraceback (most recent call last):\n')

            # Append this traceback.
            exc_full_buffer.write(strjoin.join(
                # List of lines formatted from this list.
                traceback.format_list(
                    # List of stack trace entries from this traceback.
                    traceback.extract_tb(
                        exc_parent_traceback))))

        # If this exception is *NOT* the last, append an explanatory header.
        if exc_parent_index != exc_parent_last_index:
            exc_full_buffer.write(
                '\n'
                'The above exception wrapped '
                'the following originating exception:'
                '\n\n'
            )

    # Append a random error haiku to the traceback buffer... *BECAUSE*!
    exc_full_buffer.write('\n{}'.format(errhaiku.get_random()))

    # Return the string contents of these buffers in the expected order.
    return (exc_iota_buffer.getvalue(), exc_full_buffer.getvalue())

# ....................{ PRIVATE ~ iterators               }....................
# The active Python interpreter is guaranteed to now target at least Python
# 3.6, which replaced the private traceback._iter_chain() function with a new
# public class hierarchy (e.g., "TracebackException"). For portability, we
# forward port the traceback._iter_chain() method from the most recent stable
# release of Python 3.4.
#
# Alternately, this class hierarchy *COULD* be backported from the most recent
# stable release of Python 3.6. Doing so, however, would be considerably more
# difficult than simply defining a single function. Thus the current approach.

def _iter_chain(exc, custom_tb=None, seen=None):
    '''
    Private :func:`traceback._iter_chain` method forward-ported without
    modification from the most recent stable release of Python 3.4 as of this
    writing (i.e., Python 3.4.3).
    '''

    if seen is None:
        seen = set()
    seen.add(exc)
    its = []
    context = exc.__context__
    cause = exc.__cause__
    if cause is not None and cause not in seen:
        its.append(_iter_chain(cause, False, seen))
        its.append([(_cause_message, None)])
    elif (context is not None and
        not exc.__suppress_context__ and
        context not in seen):
        its.append(_iter_chain(context, None, seen))
        its.append([(_context_message, None)])
    its.append([(exc, custom_tb or exc.__traceback__)])
    # itertools.chain is in an extension module and may be unavailable
    for it in its:
        yield from it
